# Copyright 2021 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add grant and grant request

Revision ID: 3ff50553e9d3
Revises: 70df18f71ba2
Create Date: 2021-10-07 03:57:25.430532

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '3ff50553e9d3'
down_revision = '70df18f71ba2'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.create_table('GrantV1',
        sa.Column('id', sa.String(length=255), nullable=False),
        sa.Column('vnfInstanceId', sa.String(length=255), nullable=False),
        sa.Column('vnfLcmOpOccId', sa.String(length=255), nullable=False),
        sa.Column('vimConnectionInfo', sa.JSON(), nullable=True),
        sa.Column('zones', sa.JSON(), nullable=True),
        sa.Column('zoneGroups', sa.JSON(), nullable=True),
        sa.Column('addResources', sa.JSON(), nullable=True),
        sa.Column('tempResources', sa.JSON(), nullable=True),
        sa.Column('removeResources', sa.JSON(), nullable=True),
        sa.Column('updateResources', sa.JSON(), nullable=True),
        sa.Column('vimAssets', sa.JSON(), nullable=True),
        sa.Column('extVirtualLinks', sa.JSON(), nullable=True),
        sa.Column('extManagedVirtualLinks', sa.JSON(), nullable=True),
        sa.Column('additionalParams', sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )

    op.create_table('GrantRequestV1',
        sa.Column('vnfInstanceId', sa.String(length=255), nullable=False),
        sa.Column('vnfLcmOpOccId', sa.String(length=255), nullable=False),
        sa.Column('vnfdId', sa.String(length=255), nullable=False),
        sa.Column('dstVnfdId', sa.String(length=255), nullable=True),
        sa.Column('flavourId', sa.String(length=255), nullable=True),
        sa.Column('operation',
                  sa.Enum('INSTANTIATE', 'SCALE', 'SCALE_TO_LEVEL',
                          'CHANGE_FLAVOUR', 'TERMINATE', 'HEAL', 'OPERATE',
                          'CHANGE_EXT_CONN', 'CREATE_SNAPSHOT',
                          'REVERT_TO_SNAPSHOT', 'CHANGE_VNFPKG'),
                  nullable=False),
        sa.Column('isAutomaticInvocation', sa.Boolean(), nullable=False),
        sa.Column('instantiationLevelId', sa.String(length=255),
                  nullable=True),
        sa.Column('addResources', sa.JSON(), nullable=True),
        sa.Column('tempResources', sa.JSON(), nullable=True),
        sa.Column('removeResources', sa.JSON(), nullable=True),
        sa.Column('updateResources', sa.JSON(), nullable=True),
        sa.Column('placementConstraints', sa.JSON(), nullable=True),
        sa.Column('vimConstraints', sa.JSON(), nullable=True),
        sa.Column('additionalParams', sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint('vnfLcmOpOccId'),
        mysql_engine='InnoDB'
    )



# Copyright 2018, OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Borrowed from nova code base, more utilities will be added/borrowed as and
# when needed.

"""Utilities and helper functions."""

from oslo_utils import timeutils

from tacker.db.common_services import common_services_db_plugin
from tacker.plugins.common import constants


def log_events(context, vnf_dict, evt_type, evt_details):
    _cos_db_plg = common_services_db_plugin.CommonServicesPluginDb()
    _cos_db_plg.create_event(context, res_id=vnf_dict['id'],
                             res_type=constants.RES_TYPE_VNF,
                             res_state=vnf_dict['status'],
                             evt_type=evt_type,
                             tstamp=timeutils.utcnow(),
                             details=evt_details)



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import netaddr

from mistral_lib import actions
from oslo_config import cfg
from oslo_log import log as logging

from tacker.agent.linux import utils as linux_utils
from tacker.common import rpc
from tacker.common import topics
from tacker.conductor.conductorrpc import vim_monitor_rpc
from tacker import context as t_context

LOG = logging.getLogger(__name__)


class PingVimAction(actions.Action):

    def __init__(self, count, targetip, vim_id,
                 interval, timeout):
        self.killed = False
        self.count = count
        self.timeout = timeout
        self.interval = interval
        self.targetip = targetip
        self.vim_id = vim_id
        self.current_status = "PENDING"

    def start_rpc_listeners(self):
        """Start the RPC loop to let the server communicate with actions."""
        self.endpoints = [self]
        self.conn = rpc.create_connection()
        self.conn.create_consumer(topics.TOPIC_ACTION_KILL,
                                  self.endpoints, fanout=False,
                                  host=self.vim_id)
        return self.conn.consume_in_threads()

    def killAction(self, context, **kwargs):
        self.killed = True

    def _ping(self):

        cmd_ping = 'ping'
        if netaddr.valid_ipv6(self.targetip):
            cmd_ping = 'ping6'

        ping_cmd = [cmd_ping, '-c', self.count,
                    '-W', self.timeout,
                    '-i', self.interval,
                    self.targetip]

        try:
            # NOTE(gongysh) since it is called in a loop, the debug log
            # should be disabled to avoid eating up mistral executor.
            linux_utils.execute(ping_cmd, check_exit_code=True,
                                debuglog=False)
            return 'REACHABLE'
        except RuntimeError:
            LOG.warning(("Cannot ping ip address: %s"), self.targetip)
            return 'UNREACHABLE'

    def _update(self, status):
        LOG.info("VIM %s changed to status %s", self.vim_id, status)
        target = vim_monitor_rpc.VIMUpdateRPC.target
        rpc_client = rpc.get_client(target)
        cctxt = rpc_client.prepare()
        return cctxt.call(t_context.get_admin_context_without_session(),
                          'update_vim',
                          vim_id=self.vim_id,
                          status=status)

    def run(self, action_ctx):
        servers = []
        try:
            rpc.init_action_rpc(cfg.CONF)
            servers = self.start_rpc_listeners()
        except Exception:
            LOG.exception('failed to start rpc in vim action')
            return 'FAILED'
        try:
            while True:
                if self.killed:
                    break
                status = self._ping()
                if self.current_status != status:
                    self.current_status = self._update(status)
                # TODO(gongysh) If we need to sleep a little time here?
        except Exception:
            LOG.exception('failed to run mistral action for vim %s',
                          self.vim_id)
            return 'FAILED'
        # to stop rpc connection
        for server in servers:
            try:
                server.stop()
            except Exception:
                LOG.exception(
                    'failed to stop rpc connection for vim %s',
                    self.vim_id)
        return 'KILLED'

    def test(self):
        return 'REACHABLE'



# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker.api import extensions


class VNFMonitorAbstractDriver(extensions.PluginInterface,
        metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_type(self):
        """Return one of predefined type of the hosting vnf drivers."""
        pass

    @abc.abstractmethod
    def get_name(self):
        """Return a symbolic name for the VNF Monitor plugin."""
        pass

    @abc.abstractmethod
    def get_description(self):
        """Return description of VNF Monitor plugin."""
        pass

    def monitor_get_config(self, plugin, context, vnf):
        """Return dict of monitor configuration data.

        :param plugin:
        :param context:
        :param vnf:
        :returns: dict
        :returns: dict of monitor configuration data
        """
        return {}

    @abc.abstractmethod
    def monitor_url(self, plugin, context, vnf):
        """Return the url of vnf to monitor.

        :param plugin:
        :param context:
        :param vnf:
        :returns: string
        :returns: url of vnf to monitor
        """
        pass

    @abc.abstractmethod
    def monitor_call(self, vnf, kwargs):
        """Monitor.

        Return boolean value True if VNF is healthy
        or return an event string like 'failure' or 'calls-capacity-reached'
        for specific VNF health condition.

        :param vnf:
        :param kwargs:
        :returns: boolean
        :returns: True if VNF is healthy
        """
        pass

    def monitor_service_driver(self, plugin, context, vnf,
                               service_instance):
        # use same monitor driver to communicate with service
        return self.get_name()



# Copyright (C) 2021 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ipaddress
import json
import os
import time

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils
import paramiko
import yaml

from tacker.common import cmd_executer
from tacker.common import exceptions
from tacker.common import log
from tacker.db.db_base import CommonDbMixin
from tacker.db.nfvo import nfvo_db
from tacker.extensions import nfvo
from tacker.nfvo.nfvo_plugin import NfvoPlugin
from tacker import objects
from tacker.vnfm.infra_drivers.openstack import heat_client as hc
from tacker.vnfm.mgmt_drivers import vnflcm_abstract_driver
from tacker.vnfm import vim_client

CHECK_POD_STATUS_RETRY_COUNT = 20
COMMAND_WAIT_COMPLETE_TIME = 0.2
COMMAND_WAIT_RETRY_TIME = 30
CONF = cfg.CONF
CONNECT_REMOTE_SERVER_RETRY_COUNT = 4
DRAIN_TIMEOUT = 300
K8S_CMD_TIMEOUT = 30
K8S_DEPLOY_TIMEOUT = 300
K8S_INSTALL_TIMEOUT = 2700
LOG = logging.getLogger(__name__)
NEXT_CHECK_INTERVAL_TIME = 15
ROLE_MASTER = 'master'
ROLE_WORKER = 'worker'
SERVER_WAIT_COMPLETE_TIME = 240
TOKEN_CREATE_WAIT_TIME = 30
UNINSTALL_NODE_TIMEOUT = 900


class KubesprayMgmtDriver(vnflcm_abstract_driver.VnflcmMgmtAbstractDriver):
    def get_type(self):
        return 'mgmt-drivers-kubespray'

    def get_name(self):
        return 'mgmt-drivers-kubespray'

    def get_description(self):
        return 'Tacker Kubespray VNFMgmt Driver'

    @log.log
    def instantiate_start(self, context, vnf_instance,
                          instantiate_vnf_request, grant,
                          grant_request, **kwargs):
        pass

    def _get_vim(self, context, vim_connection_info):
        vim_client_obj = vim_client.VimClient()

        if vim_connection_info:
            vim_id = vim_connection_info[0].vim_id
            access_info = vim_connection_info[0].access_info
            if access_info:
                region_name = access_info.get('region')
            else:
                region_name = None
        else:
            vim_id = None
            region_name = None

        try:
            vim_res = vim_client_obj.get_vim(
                context, vim_id, region_name=region_name)
        except nfvo.VimNotFoundException:
            raise exceptions.VimConnectionNotFound(vim_id=vim_id)

        vim_res['vim_auth'].update({'region': region_name})
        vim_info = {'id': vim_res['vim_id'], 'vim_id': vim_res['vim_id'],
                    'vim_type': vim_res['vim_type'],
                    'access_info': vim_res['vim_auth']}

        return vim_info

    def _get_vim_connection_info(self, context, instantiate_vnf_req):

        vim_info = self._get_vim(
            context, instantiate_vnf_req.vim_connection_info)

        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)

        return vim_connection_info

    def _check_is_cidr(self, key, value, cidr_str):
        # instantiate: check cidr
        try:
            ipaddress.ip_network(cidr_str)
        except ValueError:
            LOG.error('The {value} of {key} in the '
                      'additionalParams is invalid.'.format(
                          value=value, key=key))
            raise exceptions.MgmtDriverParamInvalid(param=value)

    def _check_value_exist(self, attr_list, value, key):
        for attr in attr_list:
            if not value.get(attr):
                LOG.error(
                    'The {} of {} in the '
                    'additionalParams cannot'
                    ' be None.'.format(attr, key))
                raise exceptions.MgmtDriverNotFound(
                    param=attr)

    def _check_input_parameters(self, additional_param, vnf_package_path):
        if not additional_param:
            LOG.error('The kubernetes cluster info cannot be None'
                      'in additionalParams.')
            raise exceptions.MgmtDriverOtherError(
                error_message="The kubernetes cluster info"
                              " cannot be None in additionalParams.")
        if not isinstance(additional_param, dict):
            LOG.error('The format of kubernetes cluster info in '
                      'additionalParams is invalid. It must be dict.')
            raise exceptions.MgmtDriverOtherError(
                error_message="The format of kubernetes cluster info in "
                              "additionalParams is invalid. It must be dict.")
        for key, value in additional_param.items():
            attr_list = []
            if key not in ('proxy', 'external_lb_param', 'vim_name'):
                attr_list.extend(['username', 'password'])
            if key in ('master_node', 'worker_node', 'external_lb_param'):
                attr_list.extend(['ssh_cp_name'])
            if key == 'ansible':
                attr_list.extend(['ip_address', 'kubespray_root_path',
                                 'transferring_inventory_path'])
            if key == 'external_lb_param':
                attr_list.extend(['ssh_username', 'ssh_password',
                                  'script_path'])
                if value.get('script_path'):
                    abs_script_path = os.path.join(
                        vnf_package_path, value.get('script_path'))
                    if not os.path.exists(abs_script_path):
                        LOG.error('The path of external_lb_param'
                                  ' script is invalid.')
                        raise exceptions.MgmtDriverOtherError(
                            error_message="The path of external_lb_param"
                                          " script is invalid")
            if key in ('master_node', 'ansible'):
                for attr in ['pod_cidr', 'cluster_cidr', 'ip_address']:
                    if value.get(attr):
                        self._check_is_cidr(
                            key, attr, value.get(attr))
            if attr_list:
                self._check_value_exist(attr_list, value, key)

    def _get_ssh_ip_and_nic_ip(self, heatclient, stack_id, node):
        resource_info = heatclient.resources.get(
            stack_id=stack_id,
            resource_name=node.get('ssh_cp_name'))
        if resource_info.attributes.get('floating_ip_address'):
            ssh_ip = resource_info.attributes.get('floating_ip_address')
        else:
            ssh_ip = resource_info.attributes.get(
                'fixed_ips')[0].get('ip_address')
        if not ssh_ip:
            LOG.error("Failed to get the node's ssh ip.")
            raise exceptions.MgmtDriverOtherError(
                error_message="Failed to get"
                              " the node's ssh ip.")
        if not node.get('nic_cp_name'):
            nic_ip = ssh_ip
        else:
            nic_ip = heatclient.resources.get(
                stack_id=stack_id,
                resource_name=node.get('nic_cp_name')).attributes.get(
                'fixed_ips')[0].get('ip_address')
            if not nic_ip:
                LOG.error("Failed to get the node's nic ip.")
                raise exceptions.MgmtDriverOtherError(
                    error_message="Failed to get"
                                  " the node's nic ip.")
        return ssh_ip, nic_ip

    def _get_group_resources_list(
            self, heatclient, stack_id, node, additional_params):
        # get group resources list
        nest_resources_list = heatclient.resources.list(stack_id=stack_id)
        group_stack_name = node.get("aspect_id")
        group_stack_id = ""
        for nest_resources in nest_resources_list:
            if nest_resources.resource_name == group_stack_name:
                group_stack_id = nest_resources.physical_resource_id
        if not group_stack_id:
            LOG.error('No stack id {} matching the group was found.'.format(
                group_stack_id))
            raise exceptions.MgmtDriverOtherError(
                error_message='No stack id {} matching the'
                              ' group was found.'.format(group_stack_id))
        group_resources_list = heatclient.resources.list(
            stack_id=group_stack_id)
        return group_resources_list

    def _get_install_info_for_k8s_node(self, nest_stack_id, node,
                                       additional_params, heatclient):
        # instantiate: get k8s ssh ips
        vm_dict_list = []

        # get ssh_ip and nic_ip from heat, and set value into vm_dict
        if not node.get('aspect_id'):
            ssh_ip, nic_ip = self._get_ssh_ip_and_nic_ip(
                heatclient, nest_stack_id, node)
            vm_dict = {
                "ssh_ip": ssh_ip,
                "nic_ip": nic_ip
            }
            vm_dict_list.append(vm_dict)
        else:
            group_resources_list = self._get_group_resources_list(
                heatclient, nest_stack_id, node, additional_params)
            for group_resource in group_resources_list:
                stack_id = group_resource.physical_resource_id
                ssh_ip, nic_ip = self._get_ssh_ip_and_nic_ip(
                    heatclient, stack_id, node)
                vm_dict = {
                    "ssh_ip": ssh_ip,
                    "nic_ip": nic_ip
                }
                vm_dict_list.append(vm_dict)
        return vm_dict_list

    def _set_lb_info(self, nest_stack_id, external_lb_param, master_node,
                     heatclient):
        # get ssh_ip and cluster_ip from heat, and set value into vm_dict
        ssh_ip, _ = self._get_ssh_ip_and_nic_ip(
            heatclient, nest_stack_id, external_lb_param)
        external_lb_param['pod_cidr'] = master_node.get('pod_cidr', '')
        external_lb_param['cluster_cidr'] = master_node.get(
            'cluster_cidr', '')
        external_lb_param['ssh_ip'] = ssh_ip
        external_lb_param['cluster_ip'] = ssh_ip

    def _init_commander_and_set_script(self, user, password, host,
                                       timeout, vnf_package_path=None,
                                       script_path=None, token_flag=False):
        retry = CONNECT_REMOTE_SERVER_RETRY_COUNT
        while retry > 0:
            try:
                if (vnf_package_path and script_path) or token_flag:
                    connect = paramiko.Transport(host, 22)
                    connect.connect(username=user, password=password)
                    sftp = paramiko.SFTPClient.from_transport(connect)
                    if vnf_package_path and script_path:
                        sftp.put(os.path.join(vnf_package_path, script_path),
                                 "/tmp/{}".format(
                                     script_path.replace('Scripts', '')))
                    if token_flag:
                        fname = 'create_admin_token.yaml'
                        sftp.put(os.path.join(
                            os.path.dirname(os.path.abspath(__file__)),
                            "../../../samples/mgmt_driver/{}".format(fname)),
                            "/tmp/{}".format(fname))
                    connect.close()
                commander = cmd_executer.RemoteCommandExecutor(
                    user=user, password=password, host=host,
                    timeout=timeout)
                return commander
            except (exceptions.NotAuthorized, paramiko.SSHException,
                    paramiko.ssh_exception.NoValidConnectionsError) as e:
                LOG.debug(e)
                retry -= 1
                if retry == 0:
                    LOG.error(e)
                    raise exceptions.MgmtDriverOtherError(error_message=e)
                time.sleep(SERVER_WAIT_COMPLETE_TIME)

    def _send_or_receive_file(self, host, user, password,
                              remote_file, local_file, operation):
        connect = paramiko.Transport(host, 22)
        connect.connect(username=user, password=password)
        sftp = paramiko.SFTPClient.from_transport(connect)
        if operation == 'receive':
            sftp.get(remote_file, local_file)
        else:
            sftp.put(local_file, remote_file)
        connect.close()

    def _execute_command(self, commander, ssh_command, timeout, type, retry):
        eventlet.monkey_patch()
        while retry >= 0:
            try:
                with eventlet.Timeout(timeout, True):
                    result = commander.execute_command(
                        ssh_command)
                    break
            except eventlet.timeout.Timeout:
                LOG.debug('It is time out, When execute command: '
                          '{}.'.format(ssh_command))
                retry -= 1
                if retry < 0:
                    LOG.error('It is time out, When execute command: '
                              '{}.'.format(ssh_command))
                    raise exceptions.MgmtDriverOtherError(
                        error_message='It is time out, When execute command: '
                                      '{}.'.format(ssh_command))
                time.sleep(COMMAND_WAIT_RETRY_TIME)
        if type == 'common':
            if result.get_return_code() != 0 and result.get_stderr():
                err = result.get_stderr()
                LOG.error(err)
                raise exceptions.MgmtDriverRemoteCommandError(err_info=err)
        elif type == 'ansible':
            if result.get_return_code() != 0 \
                    and 'No such file or directory' in result.get_stderr()[0]:
                return False
            else:
                error_message = 'The transferring_inventory_path has ' \
                                'exists in kubespray server. Please check' \
                                ' your path.'
                LOG.error(error_message)
                raise exceptions.MgmtDriverRemoteCommandError(
                    err_info=error_message)
        elif type == 'install':
            if result.get_return_code() != 0:
                for error in result.get_stdout():
                    if 'Timeout (12s) waiting for ' \
                        'privilege escalation prompt' in error and \
                            retry > 0:
                        self._execute_command(commander, ssh_command,
                                              timeout, 'install', 0)
                        break
                else:
                    err = result.get_stderr()
                    LOG.error(err)
                    raise exceptions.MgmtDriverRemoteCommandError(
                        err_info=err)

        return result.get_stdout()

    def _create_hosts_yaml(self, master_node, master_vm_dict_list,
                           worker_node, worker_vm_dict_list):
        hosts_yaml_content = {
            'all': {
                'hosts': {},
                'children': {
                    'kube_control_plane': {'hosts': {}},
                    'kube_node': {'hosts': {}},
                    'etcd': {'hosts': {}},
                    'k8s_cluster': {
                        'children': {'kube_control_plane': None,
                                     'kube_node': None}},
                    'calico_rr': {'hosts': {}}}}}

        for master_vm in master_vm_dict_list:
            key = 'master' + master_vm.get('nic_ip').split('.')[-1]
            hosts_yaml_content['all']['hosts'][key] = {
                'ansible_host': master_vm.get('ssh_ip'),
                'ip': master_vm.get('nic_ip'),
                'ansible_user': master_node.get('username'),
                'ansible_password': master_node.get('password'),
            }
            hosts_yaml_content['all']['children']['kube_control_plane'][
                'hosts'][key] = None
            hosts_yaml_content['all']['children']['etcd'][
                'hosts'][key] = None

        for worker_vm in worker_vm_dict_list:
            key = 'worker' + worker_vm.get('nic_ip').split('.')[-1]
            hosts_yaml_content['all']['hosts'][key] = {
                'ansible_host': worker_vm.get('ssh_ip'),
                'ip': worker_vm.get('nic_ip'),
                'ansible_user': worker_node.get('username'),
                'ansible_password': worker_node.get('password'),
            }
            hosts_yaml_content['all']['children']['kube_node'][
                'hosts'][key] = None

        return hosts_yaml_content

    def _install_k8s_cluster_and_set_config(
            self, master_node, worker_node, proxy, ansible,
            external_lb_param, master_vm_dict_list, worker_vm_dict_list):
        """Install Kubernetes Cluster Function

        It will use Kubespray which is installed in advance to install
        a Kubernetes Cluster.
        At present, Kuberspray's version is v2.16.0. You can get detailed
        information from the following url.
        https://github.com/kubernetes-sigs/kubespray/tree/v2.16.0
        """
        # get mtu value
        master_commander = self._init_commander_and_set_script(
            master_node.get('username'), master_node.get('password'),
            master_vm_dict_list[0].get('ssh_ip'), K8S_CMD_TIMEOUT)
        ssh_command = "ip a | grep '%(nic_ip)s' -B 2 | " \
                      "grep 'mtu' | awk '{print $5}'" % \
                      {'nic_ip': master_vm_dict_list[0].get('nic_ip')}
        mtu_value = self._execute_command(
            master_commander, ssh_command,
            K8S_CMD_TIMEOUT, 'common', 0)[0].replace('\n', '')
        calico_veth_mtu = int(mtu_value) - 20
        master_commander.close_session()

        # create inventory/hosts.yaml
        ansible_commander = self._init_commander_and_set_script(
            ansible.get('username'), ansible.get('password'),
            ansible.get('ip_address'), K8S_CMD_TIMEOUT)
        ssh_command = "ls -l {}".format(
            ansible.get('transferring_inventory_path'))
        file_exists_flag = self._execute_command(
            ansible_commander, ssh_command, K8S_CMD_TIMEOUT, 'ansible', 0)
        if not file_exists_flag:
            ssh_command = 'cp -r {kubespray_root_path}/inventory/sample' \
                          ' {transferring_inventory_path}'.format(
                              kubespray_root_path=ansible.get(
                                  'kubespray_root_path'),
                              transferring_inventory_path=ansible.get(
                                  'transferring_inventory_path'))
            self._execute_command(
                ansible_commander, ssh_command, K8S_CMD_TIMEOUT, 'common', 0)
        hosts_yaml_content = self._create_hosts_yaml(
            master_node, master_vm_dict_list,
            worker_node, worker_vm_dict_list)
        local_hosts_yaml_path = '/tmp/hosts.yaml'
        with open(local_hosts_yaml_path, 'w', encoding='utf-8') as nf:
            yaml.safe_dump(hosts_yaml_content, nf, default_flow_style=False)
        remote_hosts_yaml_path = ansible.get(
            'transferring_inventory_path') + '/hosts.yaml'
        self._send_or_receive_file(
            ansible.get('ip_address'), ansible.get('username'),
            ansible.get('password'), remote_hosts_yaml_path,
            local_hosts_yaml_path, 'send')

        # set calico mtu value
        calico_file_path = ansible.get(
            'transferring_inventory_path') + \
            '/group_vars/k8s_cluster/k8s-net-calico.yml'
        ssh_command = 'sed -i "s/\\# calico_mtu: 1500/calico_mtu: ' \
                      '{mtu_value}/g" {calico_file_path}'.format(
                          mtu_value=mtu_value,
                          calico_file_path=calico_file_path)
        self._execute_command(ansible_commander, ssh_command,
                              K8S_CMD_TIMEOUT, 'common', 0)
        ssh_command = 'sed -i "s/\\# calico_veth_mtu: 1440/calico_veth_mtu:' \
                      ' {calico_veth_mtu}/g" {calico_file_path}'.format(
                          calico_veth_mtu=calico_veth_mtu,
                          calico_file_path=calico_file_path)
        self._execute_command(ansible_commander, ssh_command,
                              K8S_CMD_TIMEOUT, 'common', 0)

        # set pod and service cidr information
        if external_lb_param.get('cluster_cidr') and \
                external_lb_param.get('pod_cidr'):
            k8s_cluster_file_path = ansible.get(
                'transferring_inventory_path') + \
                '/group_vars/k8s_cluster/k8s-cluster.yml'
            cluster_cidr = external_lb_param.get(
                'cluster_cidr').replace('/', '\\/')
            ssh_command = 'sed -i "s/kube_service_addresses:' \
                          ' 10.233.0.0\\/18/' \
                          'kube_service_addresses: {k8s_service_address}/g"' \
                          ' {k8s_cluster_file_path}'.format(
                              k8s_service_address=cluster_cidr,
                              k8s_cluster_file_path=k8s_cluster_file_path)
            self._execute_command(ansible_commander, ssh_command,
                                  K8S_CMD_TIMEOUT, 'common', 0)
            pod_cidr = external_lb_param.get('pod_cidr').replace('/', '\\/')
            ssh_command = 'sed -i "s/kube_pods_subnet: 10.233.64.0\\/18/' \
                          'kube_pods_subnet: {pod_cidr}/g"' \
                          ' {k8s_cluster_file_path}'.format(
                              pod_cidr=pod_cidr,
                              k8s_cluster_file_path=k8s_cluster_file_path)
            self._execute_command(ansible_commander, ssh_command,
                                  K8S_CMD_TIMEOUT, 'common', 0)

        # set proxy
        if proxy:
            proxy_file_path = ansible.get(
                'transferring_inventory_path') + \
                '/group_vars/all/all.yml'
            http_proxy = proxy.get('http_proxy').replace('/', '\\/')
            https_proxy = proxy.get('http_proxy').replace('/', '\\/')
            ssh_command = 'sed -i "s/\\# http_proxy: \\"\\"/' \
                          'http_proxy: {http_proxy}/g"' \
                          ' {proxy_file_path}'.format(
                              http_proxy=http_proxy,
                              proxy_file_path=proxy_file_path)
            self._execute_command(ansible_commander, ssh_command,
                                  K8S_CMD_TIMEOUT, 'common', 0)
            ssh_command = 'sed -i "s/\\# https_proxy: \\"\\"/' \
                          'https_proxy: {https_proxy}/g"' \
                          ' {proxy_file_path}'.format(
                              https_proxy=https_proxy,
                              proxy_file_path=proxy_file_path)
            self._execute_command(ansible_commander, ssh_command,
                                  K8S_CMD_TIMEOUT, 'common', 0)
        ansible_commander.close_session()

        # install k8s cluster
        install_timeout = K8S_INSTALL_TIMEOUT * (
            len(master_vm_dict_list) + len(worker_vm_dict_list))
        ansible_commander = self._init_commander_and_set_script(
            ansible.get('username'), ansible.get('password'),
            ansible.get('ip_address'), install_timeout)
        cluster_yaml_path = ansible.get(
            'kubespray_root_path') + '/cluster.yml'
        ssh_command = 'ansible-playbook -i {}/hosts.yaml --become' \
                      ' --become-user=root {}'.format(
                          ansible.get('transferring_inventory_path'),
                          cluster_yaml_path)
        self._execute_command(ansible_commander, ssh_command,
                              install_timeout, 'install', 1)
        ansible_commander.close_session()

        # get k8s bearer token
        master_commander = self._init_commander_and_set_script(
            master_node.get('username'), master_node.get('password'),
            master_vm_dict_list[0].get('ssh_ip'), K8S_CMD_TIMEOUT,
            token_flag=True)
        ssh_command = "sudo kubectl create -f /tmp/create_admin_token.yaml"
        self._execute_command(
            master_commander, ssh_command, K8S_CMD_TIMEOUT, 'common', 0)
        time.sleep(TOKEN_CREATE_WAIT_TIME)
        ssh_command = "sudo kubectl get secret -n kube-system " \
                      "| grep '^admin-token' " \
                      "| awk '{print $1}' " \
                      "| xargs -i sudo kubectl get secret {} -n kube-system" \
                      " -ojsonpath={.data.token} | base64 -d"
        bearer_token = self._execute_command(
            master_commander, ssh_command,
            K8S_CMD_TIMEOUT, 'common', 0)[0].replace('\n', '')
        master_commander.close_session()
        if os.path.exists(local_hosts_yaml_path):
            os.remove(local_hosts_yaml_path)

        return bearer_token

    def _install_and_set_lb(self, external_lb_param, vnf_package_path, proxy,
                            master_vm_dict_list, worker_vm_dict_list,
                            master_node):
        lb_commander = self._init_commander_and_set_script(
            external_lb_param.get('ssh_username'),
            external_lb_param.get('ssh_password'),
            external_lb_param.get('ssh_ip'), K8S_DEPLOY_TIMEOUT,
            vnf_package_path=vnf_package_path,
            script_path=external_lb_param.get('script_path'))
        master_ssh_ips_str = ','.join([vm_dict.get(
            'nic_ip') for vm_dict in master_vm_dict_list])
        worker_ssh_ips_str = ','.join([vm_dict.get(
            'nic_ip') for vm_dict in worker_vm_dict_list])
        if proxy.get('http_proxy') and proxy.get('https_proxy'):
            ssh_command = \
                "export http_proxy={http_proxy};" \
                "export https_proxy={https_proxy};" \
                "bash /tmp/{script_path} " \
                "-m {master_ip} -w {worker_ip} ".format(
                    http_proxy=proxy.get('http_proxy'),
                    https_proxy=proxy.get('https_proxy'),
                    master_ip=master_ssh_ips_str,
                    worker_ip=worker_ssh_ips_str,
                    script_path=external_lb_param.get(
                        'script_path').replace('Scripts/', ''))
        else:
            ssh_command = \
                "bash /tmp/{script_path} " \
                "-m {master_ip} -w {worker_ip} ".format(
                    master_ip=master_ssh_ips_str,
                    worker_ip=worker_ssh_ips_str,
                    script_path=external_lb_param.get(
                        'script_path').replace('Scripts/', ''))
        self._execute_command(
            lb_commander, ssh_command, K8S_DEPLOY_TIMEOUT, 'common', 0)
        lb_commander.close_session()

        # copy k8s admin configuration file
        master_commander = self._init_commander_and_set_script(
            master_node.get('username'), master_node.get('password'),
            master_vm_dict_list[0].get('ssh_ip'), K8S_CMD_TIMEOUT)
        ssh_command = 'sudo cp /etc/kubernetes/admin.conf /tmp/config;' \
                      'sudo chown $(id -u):$(id -g) /tmp/config'
        self._execute_command(master_commander, ssh_command,
                              K8S_CMD_TIMEOUT, 'common', 0)
        ssh_command = "sed -i 's/:6443/:8383/' /tmp/config"
        self._execute_command(master_commander, ssh_command,
                              K8S_CMD_TIMEOUT, 'common', 0)
        master_commander.close_session()
        remote_admin_file_path = local_admin_file_path = '/tmp/config'
        self._send_or_receive_file(
            master_vm_dict_list[0].get('ssh_ip'),
            master_node.get('username'), master_node.get('password'),
            remote_admin_file_path, local_admin_file_path, 'receive')

        # send config file to lb server
        lb_admin_file_path = '~/.kube/config'
        if os.path.exists(local_admin_file_path):
            self._send_or_receive_file(
                external_lb_param.get('ssh_ip'),
                external_lb_param.get('ssh_username'),
                external_lb_param.get('ssh_password'),
                remote_admin_file_path, local_admin_file_path, 'send')
        lb_commander = self._init_commander_and_set_script(
            external_lb_param.get('ssh_username'),
            external_lb_param.get('ssh_password'),
            external_lb_param.get('ssh_ip'), K8S_CMD_TIMEOUT)
        ssh_command = "mv {} {}".format(remote_admin_file_path,
                                        lb_admin_file_path)
        self._execute_command(lb_commander, ssh_command,
                              K8S_CMD_TIMEOUT, 'common', 0)
        lb_commander.close_session()

        if os.path.exists(local_admin_file_path):
            os.remove(local_admin_file_path)

    def _create_vim(self, context, vnf_instance, external_lb_param,
                    bearer_token, vim_name):
        server = 'https://' + external_lb_param.get('cluster_ip') + ':8383'
        vim_info = {
            'vim': {
                'name': vim_name,
                'auth_url': server,
                'vim_project': {
                    'name': 'default'
                },
                'auth_cred': {
                    'bearer_token': bearer_token
                },
                'type': 'kubernetes',
                'tenant_id': context.project_id
            }
        }
        try:
            nfvo_plugin = NfvoPlugin()
            created_vim_info = nfvo_plugin.create_vim(context, vim_info)
        except Exception as e:
            LOG.error("Failed to register kubernetes vim: {}".format(e))
            raise exceptions.MgmtDriverOtherError(
                error_message="Failed to register kubernetes vim: {}".format(
                    e))
        id = uuidutils.generate_uuid()
        vim_id = created_vim_info.get('id')
        vim_type = 'kubernetes'
        access_info = {
            'auth_url': server
        }
        vim_connection_info = objects.VimConnectionInfo(
            id=id, vim_id=vim_id, vim_type=vim_type,
            access_info=access_info, interface_info=None
        )
        vim_connection_infos = vnf_instance.vim_connection_info
        vim_connection_infos.append(vim_connection_info)
        vnf_instance.vim_connection_info = vim_connection_infos
        vnf_instance.save()

    def _get_vnf_package_path(self, context, vnfd_id):
        return os.path.join(CONF.vnf_package.vnf_package_csar_path,
                            self._get_vnf_package_id(context, vnfd_id))

    def _get_vnf_package_id(self, context, vnfd_id):
        vnf_package = objects.VnfPackageVnfd.get_by_id(context, vnfd_id)
        return vnf_package.package_uuid

    @log.log
    def instantiate_end(self, context, vnf_instance,
                        instantiate_vnf_request, grant,
                        grant_request, **kwargs):
        # get vim_connect_info
        if hasattr(instantiate_vnf_request, 'vim_connection_info'):
            vim_connection_info = self._get_vim_connection_info(
                context, instantiate_vnf_request)
        else:
            # In case of healing entire Kubernetes cluster, 'heal_end' method
            # will call this method using 'vnf_instance.instantiated_vnf_info'
            # as the 'instantiate_vnf_request', but there is no
            # 'vim_connection_info' in it, so we should get
            # 'vim_connection_info' from 'vnf_instance'.
            vim_connection_info = self._get_vim_connection_info(
                context, vnf_instance)
        additional_param = instantiate_vnf_request.additional_params.get(
            'k8s_cluster_installation_param', {})
        vim_name = additional_param.get('vim_name')
        master_node = additional_param.get('master_node', {})
        worker_node = additional_param.get('worker_node', {})
        proxy = additional_param.get('proxy', {})
        ansible = additional_param.get('ansible', {})
        external_lb_param = additional_param.get('external_lb_param', {})
        vnf_package_path = self._get_vnf_package_path(
            context, vnf_instance.vnfd_id)
        self._check_input_parameters(additional_param, vnf_package_path)
        nest_stack_id = vnf_instance.instantiated_vnf_info.instance_id
        if not vim_name:
            vim_name = 'kubernetes_vim_' + vnf_instance.id

        # get k8s node vm list
        access_info = vim_connection_info.access_info
        heatclient = hc.HeatClient(access_info)
        master_vm_dict_list = \
            self._get_install_info_for_k8s_node(
                nest_stack_id, master_node,
                instantiate_vnf_request.additional_params,
                heatclient)
        worker_vm_dict_list = \
            self._get_install_info_for_k8s_node(
                nest_stack_id, worker_node,
                instantiate_vnf_request.additional_params, heatclient)

        # set LB vm's info
        self._set_lb_info(nest_stack_id, external_lb_param, master_node,
                          heatclient)

        # install k8s_cluster and set config
        bearer_token = self._install_k8s_cluster_and_set_config(
            master_node, worker_node, proxy, ansible, external_lb_param,
            master_vm_dict_list, worker_vm_dict_list)

        # Install and set ExternalLB
        self._install_and_set_lb(external_lb_param, vnf_package_path, proxy,
                                 master_vm_dict_list, worker_vm_dict_list,
                                 master_node)

        # create vim
        self._create_vim(context, vnf_instance, external_lb_param,
                         bearer_token, vim_name)

    @log.log
    def terminate_start(self, context, vnf_instance,
                        terminate_vnf_request, grant,
                        grant_request, **kwargs):
        pass

    def _get_vim_by_name(self, context, k8s_vim_name):
        common_db_api = CommonDbMixin()
        result = common_db_api.get_by_name(
            context, nfvo_db.Vim, k8s_vim_name)

        if not result:
            LOG.debug("Cannot find kubernetes "
                      "vim with name: {}".format(k8s_vim_name))

        return result

    @log.log
    def terminate_end(self, context, vnf_instance,
                      terminate_vnf_request, grant,
                      grant_request, **kwargs):
        # delete kubernetes vim
        k8s_params = vnf_instance.instantiated_vnf_info.additional_params.get(
            'k8s_cluster_installation_param', {})
        k8s_vim_name = k8s_params.get('vim_name')
        if not k8s_vim_name:
            k8s_vim_name = 'kubernetes_vim_' + vnf_instance.id

        vim_info = self._get_vim_by_name(
            context, k8s_vim_name)
        if vim_info:
            nfvo_plugin = NfvoPlugin()
            nfvo_plugin.delete_vim(context, vim_info.id)
            for k8s_vim in vnf_instance.vim_connection_info:
                if k8s_vim.vim_id == vim_info.id:
                    vnf_instance.vim_connection_info.remove(k8s_vim)

        # delete cluster info on ansible server
        _, _, ansible, _ = \
            self._get_initial_parameters(
                context, vnf_instance, terminate_vnf_request)
        commander = self._init_commander_and_set_script(
            ansible.get('username'), ansible.get('password'),
            ansible.get('ip_address'), K8S_CMD_TIMEOUT)
        ssh_command = 'rm -rf {}'.format(
            k8s_params.get('ansible').get('transferring_inventory_path'))
        self._execute_command(
            commander, ssh_command, K8S_CMD_TIMEOUT, 'common', 0)
        ssh_command = 'rm -rf ~/.ssh/known_hosts'
        self._execute_command(
            commander, ssh_command, K8S_CMD_TIMEOUT, 'common', 0)
        commander.close_session()

    def _update_external_lb(self, external_lb_param, lb_ssh_ip, hostname):
        external_lb_commander = self._init_commander_and_set_script(
            external_lb_param.get('ssh_username'),
            external_lb_param.get('ssh_password'),
            lb_ssh_ip,
            K8S_CMD_TIMEOUT
        )
        ssh_command = 'grep -n "{}" /etc/haproxy/haproxy.cfg | ' \
                      'cut -d : -f 1'.format(hostname)
        result = self._execute_command(
            external_lb_commander, ssh_command,
            K8S_CMD_TIMEOUT, 'common', 0)
        if result:
            worker_host_num = result[0].replace('\n', '')
            if worker_host_num.isdigit():
                lb_server_num = int(worker_host_num, base=0)
                ssh_command = "sudo sed -i '{}d' " \
                              "/etc/haproxy/haproxy.cfg" \
                    .format(lb_server_num)
                self._execute_command(
                    external_lb_commander, ssh_command,
                    K8S_CMD_TIMEOUT, 'common', 0)
                self._restart_haproxy(external_lb_commander)
        external_lb_commander.close_session()

    def _delete_worker_node_and_update_inventory_file(
            self, ansible, worker_node, worker_hostname, operation_type):
        update_hosts_yaml_path = ansible.get(
            'transferring_inventory_path') + '/hosts.yaml'
        self._modify_ansible_user_or_password(update_hosts_yaml_path,
                                              worker_node, ansible)
        # remove worker node
        ssh_command = "ansible-playbook -i" \
                      " {}/hosts.yaml" \
                      " --become --become-user=root " \
                      "{}/remove-node.yml -e" \
                      " node={}".format(ansible.get(
                          'transferring_inventory_path'),
                          ansible.get('kubespray_root_path'),
                          worker_hostname)
        try:
            with eventlet.Timeout(K8S_INSTALL_TIMEOUT, True):
                result, code = self._uninstall_worker_node(
                    ssh_command, ansible)
                if code != 0:
                    msg = 'Fail to remove the worker node {}'.\
                        format(worker_hostname)
                    LOG.error(result)
                    raise exceptions.MgmtDriverOtherError(
                        error_message=msg)
                LOG.debug(result)
        except eventlet.timeout.Timeout:
            msg = 'It is time out while deleting' \
                  ' the worker node {}'.format(worker_hostname)
            LOG.error(msg)
            raise exceptions.MgmtDriverOtherError(
                error_message=msg)

        # Gets the line of rows where worker_hostname resides
        if operation_type == 'SCALE':
            while True:
                commander_k8s = self._init_commander_and_set_script(
                    ansible.get('username'), ansible.get('password'),
                    ansible.get('ip_address'), K8S_CMD_TIMEOUT)
                ssh_command = 'grep -n "{}" {} | head -1 ' \
                              '| cut -d : -f 1'\
                    .format(worker_hostname, update_hosts_yaml_path)
                host_name = self._execute_command(
                    commander_k8s, ssh_command,
                    K8S_CMD_TIMEOUT, 'common', 0)
                if host_name:
                    host_name_line = host_name[0].replace('\n', '')
                    if host_name_line.isdigit():
                        host_name_line = int(host_name_line, base=0)
                        ssh_command = 'sed -n {}P {}' \
                            .format(host_name_line + 1,
                                    update_hosts_yaml_path)
                        is_hosts_or_children = self._execute_command(
                            commander_k8s, ssh_command,
                            K8S_CMD_TIMEOUT, 'common', 0)[0]
                        if "ansible_host" in is_hosts_or_children:
                            ssh_command = "sed -i '{}, {}d' {}" \
                                .format(host_name_line,
                                        host_name_line + 4,
                                        update_hosts_yaml_path)
                        else:
                            ssh_command = "sed -i " \
                                          "'{}d' {}"\
                                .format(host_name_line,
                                        update_hosts_yaml_path)
                        self._execute_command(
                            commander_k8s, ssh_command,
                            K8S_CMD_TIMEOUT, 'common', 0)
                else:
                    break
            commander_k8s.close_session()
        if os.path.exists(update_hosts_yaml_path):
            os.remove(update_hosts_yaml_path)

    def _uninstall_worker_node(self, ssh_command, ansible):
        end_str = ('# ', '$ ', '? ', '% ')
        end_flag = False
        result_end_flag = False
        command_return_code = 0
        try:
            trans = paramiko.Transport((ansible.get('ip_address'), 22))
            trans.start_client()
            trans.auth_password(username=ansible.get('username'),
                                password=ansible.get('password'))
            channel = trans.open_session()
            channel.settimeout(UNINSTALL_NODE_TIMEOUT)
            channel.get_pty()
            buff = ''
            channel.invoke_shell()
            channel.send(ssh_command + '\n')
            while True:
                time.sleep(COMMAND_WAIT_COMPLETE_TIME)
                resp = channel.recv(1024)
                resp = resp.decode('utf-8')
                buff += resp
                if "Type 'yes' to delete nodes" in resp:
                    channel.send('yes\n')
                    time.sleep(COMMAND_WAIT_COMPLETE_TIME)
                    resp = channel.recv(1024)
                    resp = resp.decode('utf-8')
                    buff += resp
                for end_s in end_str:
                    if resp.endswith(end_s):
                        end_flag = True
                        break
                if end_flag:
                    break
                if 'PLAY RECAP' in resp:
                    result_end_flag = True
                if result_end_flag and 'failed=0' not in resp:
                    command_return_code = 2
            channel.close()
            trans.close()
            return buff, command_return_code
        except (exceptions.NotAuthorized, paramiko.SSHException,
                paramiko.ssh_exception.NoValidConnectionsError) as e:
            LOG.debug(e)
            raise exceptions.MgmtDriverOtherError(error_message=e)

    def _get_initial_parameters(self, context, vnf_instance, action_request):
        vim_connection_info = \
            self._get_vim_connection_info(context, vnf_instance)
        k8s_cluster_installation_param = \
            vnf_instance.instantiated_vnf_info.additional_params.get(
                'k8s_cluster_installation_param')
        worker_node_default = \
            k8s_cluster_installation_param.get('worker_node')
        external_lb_param_default = \
            k8s_cluster_installation_param.get('external_lb_param')
        ansible_default = \
            k8s_cluster_installation_param.get('ansible')

        # If additional_params exist in action_request
        if hasattr(action_request, 'additional_params') and \
                action_request.additional_params:
            # Get the VM's information from action_request
            add_param = action_request. \
                additional_params.get('k8s_cluster_installation_param')
            if add_param:
                worker_node = add_param.get('worker_node', worker_node_default)
                external_lb_param = add_param.get('external_lb_param',
                                                  external_lb_param_default)
                ansible = add_param.get('ansible', ansible_default)
            else:
                worker_node = worker_node_default
                external_lb_param = external_lb_param_default
                ansible = ansible_default
        else:
            worker_node = worker_node_default
            external_lb_param = external_lb_param_default
            ansible = ansible_default

        return worker_node, external_lb_param, ansible, vim_connection_info

    def _remove_node_and_update_config_file(
            self, worker_hostnames, external_lb_param,
            lb_ssh_ip, ansible, worker_node, operation_type):
        # Migrate the pod of the worker node
        for worker_hostname in worker_hostnames:
            #  init lb RemoteCommandExecutor
            external_lb_commander = self._init_commander_and_set_script(
                external_lb_param.get('ssh_username'),
                external_lb_param.get('ssh_password'),
                lb_ssh_ip,
                K8S_CMD_TIMEOUT
            )

            # check worker_node exist in k8s-cluster
            ssh_command = "kubectl get node --no-headers {}" \
                          " 2> /dev/null".format(worker_hostname)
            result = self._execute_command(external_lb_commander,
                                           ssh_command,
                                           K8S_CMD_TIMEOUT,
                                           'common',
                                           0)
            if result:
                ssh_command = \
                    "kubectl get pods --field-selector=spec." \
                    "nodeName={} -o json".format(worker_hostname)
                result = self._execute_command(external_lb_commander,
                                               ssh_command,
                                               K8S_CMD_TIMEOUT,
                                               'common',
                                               0)

                # Get the names of all pods on the worker node
                daemonset_content_str = ''.join(result)
                daemonset_content = json.loads(
                    daemonset_content_str)
                ssh_command = "kubectl drain {}" \
                              " --ignore-daemonsets" \
                              " --delete-emptydir-data" \
                              " --timeout={}s".format(worker_hostname,
                                                      DRAIN_TIMEOUT)
                self._execute_command(external_lb_commander,
                                      ssh_command,
                                      K8S_DEPLOY_TIMEOUT,
                                      'common', 0)
                self.evacuate_wait(external_lb_commander,
                                   daemonset_content)
                external_lb_commander.close_session()

                # Uninstall worker node and update inventory file
                self._delete_worker_node_and_update_inventory_file(
                    ansible, worker_node, worker_hostname, operation_type)

            # Update ExternalLB's haproxy
            self._update_external_lb(external_lb_param, lb_ssh_ip,
                                     worker_hostname)
        ansible_commander = self._init_commander_and_set_script(
            ansible.get('username'), ansible.get('password'),
            ansible.get('ip_address'), K8S_CMD_TIMEOUT)
        ssh_command = 'rm -rf ~/.ssh/known_hosts'
        self._execute_command(
            ansible_commander, ssh_command, K8S_CMD_TIMEOUT, 'common', 0)
        ansible_commander.close_session()

    @log.log
    def scale_start(self, context, vnf_instance,
                    scale_vnf_request, grant,
                    grant_request, **kwargs):
        # If the type of scale is SCALE_IN
        if scale_vnf_request.type == 'SCALE_IN':
            scale_name_list = kwargs.get('scale_name_list')
            nest_stack_id = vnf_instance.instantiated_vnf_info.instance_id
            resource_name = scale_vnf_request.aspect_id

            worker_node, external_lb_param, ansible, vim_connection_info = \
                self._get_initial_parameters(
                    context, vnf_instance, scale_vnf_request)

            # Get the ssh ip of LB
            heatclient = hc.HeatClient(vim_connection_info.access_info)
            resource_info = heatclient.resources. \
                get(stack_id=nest_stack_id,
                    resource_name=external_lb_param.get('ssh_cp_name'))
            # If the VM's floating ip is not None
            # Get floating ip from resource_info and assign it to ssh ip
            lb_ssh_ip = self._get_lb_or_worker_ssh_ip(resource_info, True)

            # Get the ip of scale in worker nodes
            worker_group_resource = heatclient.resources. \
                get(stack_id=nest_stack_id,
                    resource_name=resource_name)
            # if worker_group_resource is None
            if not worker_group_resource:
                LOG.error("The specified resource was not found.")
                raise exceptions.MgmtDriverOtherError(
                    error_message='The specified resource was not found.')
            worker_resource_list = \
                heatclient.resources.list(
                    stack_id=worker_group_resource.physical_resource_id)

            worker_ip_dict_list = []
            for worker_resource in worker_resource_list:
                # If worker_resource.resource_name exists in scale_name_list
                if worker_resource.resource_name in scale_name_list:
                    stack_id = worker_resource.physical_resource_id
                    # Get the ssh_ip, nic ip of worker node
                    worker_ssh_ip, worker_nic_ip = self._get_ssh_ip_and_nic_ip(
                        heatclient, stack_id, worker_node)

                    # Create worker_ip_dict_list data
                    ip_dict = {"ssh_ip": worker_ssh_ip,
                               "nic_ip": worker_nic_ip}
                    worker_ip_dict_list.append(ip_dict)

            # Get the hostname of the scale in worker node.
            worker_hostnames = []
            for worker_ip_dict in worker_ip_dict_list:
                # get worker host names
                worker_hostname = \
                    'worker' + worker_ip_dict.get("nic_ip").split('.')[-1]
                worker_hostnames.append(worker_hostname)

            self._remove_node_and_update_config_file(
                worker_hostnames, external_lb_param,
                lb_ssh_ip, ansible, worker_node, 'SCALE')
        else:
            pass

    def evacuate_wait(self, commander, daemonset_content):
        wait_flag = True
        retry_count = CHECK_POD_STATUS_RETRY_COUNT
        while wait_flag and retry_count > 0:
            if daemonset_content.get('items'):
                ssh_command = "kubectl get pods --all-namespaces -o json"
                result = self._execute_command(
                    commander, ssh_command, K8S_CMD_TIMEOUT, 'common', 3)
                pods_list = json.loads(''.join(result)).get('items')
                pods_names = [pod.get('metadata', {}).get('name')
                              for pod in pods_list]
                for daemonset in daemonset_content.get('items'):
                    daemonset_name = daemonset.get('metadata', {}).get('name')
                    if daemonset_name in pods_names and \
                            'calico-node' not in daemonset_name and \
                            'kube-proxy' not in daemonset_name:
                        break
                else:
                    wait_flag = False
            else:
                break
            if not wait_flag:
                break
            time.sleep(NEXT_CHECK_INTERVAL_TIME)
            retry_count -= 1

    def _get_lb_or_worker_ssh_ip(self, resource_info, is_lb):
        if resource_info.attributes.get('floating_ip_address'):
            ssh_ip = resource_info.attributes.get('floating_ip_address')
        else:
            ssh_ip = resource_info.attributes. \
                get('fixed_ips')[0].get('ip_address')

        if ssh_ip is None:
            if is_lb:
                LOG.error("Failed to get the LB's ssh ip.")
                raise exceptions.MgmtDriverOtherError(
                    error_message="Failed to get the LB's ssh ip.")
            LOG.error("Failed to get the Worker's ssh ip.")
            raise exceptions.MgmtDriverOtherError(
                error_message="Failed to get the Worker's ssh ip.")
        return ssh_ip

    def _restart_haproxy(self, external_lb_commander):
        # restart haproxy
        ssh_command = 'sudo systemctl restart haproxy'
        self._execute_command(
            external_lb_commander, ssh_command,
            K8S_CMD_TIMEOUT, 'common', 0)
        ssh_command = 'sudo systemctl status haproxy | ' \
                      'grep Active'
        self._execute_command(
            external_lb_commander, ssh_command,
            K8S_CMD_TIMEOUT, 'common', 0)

    def _update_lb_config_file(self, external_lb_param, lb_ssh_ip,
                               worker_ip_dict_list):
        external_lb_commander = self._init_commander_and_set_script(
            external_lb_param.get('ssh_username'),
            external_lb_param.get('ssh_password'),
            lb_ssh_ip,
            K8S_CMD_TIMEOUT
        )
        add_row_data = ''
        for worker_ip_dict in worker_ip_dict_list:
            worker_host_name = 'worker' + \
                               worker_ip_dict.get('nic_ip').split('.')[-1]
            nic_ip = worker_ip_dict.get('nic_ip')
            row_data = '    server  {} {} check'.format(
                worker_host_name, nic_ip)
            add_row_data += row_data + '\\n'

        ssh_command = 'grep -n "backend kubernetes-nodeport" ' \
                      '/etc/haproxy/haproxy.cfg | head -1 | cut -d : -f 1'
        result = self._execute_command(external_lb_commander,
                                       ssh_command,
                                       K8S_INSTALL_TIMEOUT,
                                       'common', 0)[0].replace('\n', '')
        write_start_row = int(result) + 2
        ssh_command = 'sudo sed -i "{}a\\{}" ' \
                      '/etc/haproxy/haproxy.cfg'.format(
                          write_start_row, add_row_data)
        LOG.debug("ssh_command: {}".format(ssh_command))
        self._execute_command(
            external_lb_commander, ssh_command,
            K8S_INSTALL_TIMEOUT, 'common', 0)

        self._restart_haproxy(external_lb_commander)
        external_lb_commander.close_session()

    def _install_node_and_update_config_file(
            self, worker_node, worker_ip_dict_list,
            ansible, external_lb_param, lb_ssh_ip):
        # check worker_VM can be accessed via ssh
        self._init_commander_and_set_script(
            worker_node.get('username'), worker_node.get('password'),
            worker_ip_dict_list[0].get('ssh_ip'), K8S_CMD_TIMEOUT)

        # Install worker node
        commander_k8s = self._init_commander_and_set_script(
            ansible.get('username'), ansible.get('password'),
            ansible.get('ip_address'),
            K8S_INSTALL_TIMEOUT * len(worker_ip_dict_list))
        facts_yaml_path = ansible.get(
            'kubespray_root_path') + '/facts.yml'
        ssh_command = "ansible-playbook -i" \
                      " {}/hosts.yaml" \
                      " --become --become-user=root {}" \
            .format(ansible.get('transferring_inventory_path'),
                    facts_yaml_path)
        self._execute_command(
            commander_k8s, ssh_command,
            K8S_DEPLOY_TIMEOUT, 'common', 0)

        scale_yaml_path = ansible.get(
            'kubespray_root_path') + '/scale.yml'
        ssh_command = "ansible-playbook -i" \
                      " {}/hosts.yaml" \
                      " --become --become-user=root {}".format(
                          ansible.get('transferring_inventory_path'),
                          scale_yaml_path)
        self._execute_command(
            commander_k8s, ssh_command,
            K8S_INSTALL_TIMEOUT * len(worker_ip_dict_list),
            'install', 0)
        commander_k8s.close_session()

        # Update ExternalLB's haproxy.cfg
        self._update_lb_config_file(
            external_lb_param, lb_ssh_ip, worker_ip_dict_list)

    @log.log
    def scale_end(self, context, vnf_instance,
                  scale_vnf_request, grant,
                  grant_request, **kwargs):
        if scale_vnf_request.type == 'SCALE_OUT':
            scale_out_id_list = kwargs.get('scale_out_id_list')
            nest_stack_id = vnf_instance.instantiated_vnf_info.instance_id
            worker_node, external_lb_param, ansible, vim_connection_info =\
                self._get_initial_parameters(
                    context, vnf_instance, scale_vnf_request)

            heatclient = hc.HeatClient(vim_connection_info.access_info)

            # Get the ssh ip of LB
            resource_info = heatclient.resources. \
                get(stack_id=nest_stack_id,
                    resource_name=external_lb_param.get('ssh_cp_name'))
            lb_ssh_ip = self._get_lb_or_worker_ssh_ip(resource_info, True)

            # get scale-out worker's info
            worker_ip_dict_list = []
            for scale_out_id in scale_out_id_list:
                stack_id = scale_out_id
                # Get the ssh_ip, nic ip of worker node
                worker_ssh_ip, worker_nic_ip = self._get_ssh_ip_and_nic_ip(
                    heatclient, stack_id, worker_node)

                # Create worker_ip_dict_list data
                ip_dict = {"ssh_ip": worker_ssh_ip, "nic_ip": worker_nic_ip}
                worker_ip_dict_list.append(ip_dict)

            # read hosts.yaml file contents
            update_hosts_yaml_path = ansible.get(
                'transferring_inventory_path') + '/hosts.yaml'
            local_hosts_yaml_path = '/tmp/hosts.yaml'
            # update hosts.yaml
            hosts_content = self._modify_ansible_user_or_password(
                update_hosts_yaml_path, worker_node, ansible)

            for worker_ip_dict in worker_ip_dict_list:
                # Update inventory file
                # update hosts.yaml file contents
                worker_host_name = 'worker' + \
                                   worker_ip_dict.get('nic_ip').split('.')[-1]
                hosts_content['all']['hosts'][worker_host_name] = {
                    'ansible_host': worker_ip_dict.get('ssh_ip'),
                    'ip': worker_ip_dict.get('nic_ip'),
                    'ansible_user': worker_node.get('username'),
                    'ansible_password': worker_node.get('password')
                }
                hosts_content['all']['children']['kube_node'][
                    'hosts'][worker_host_name] = None
            LOG.debug("get hosts_content: {}".format(hosts_content))
            with open(local_hosts_yaml_path, 'w', encoding='utf-8') as nf:
                yaml.safe_dump(hosts_content, nf,
                               default_flow_style=False)
            self._send_or_receive_file(
                ansible.get('ip_address'), ansible.get('username'),
                ansible.get('password'), update_hosts_yaml_path,
                local_hosts_yaml_path, 'send')

            # Install worker node adn update configuration file
            self._install_node_and_update_config_file(
                worker_node, worker_ip_dict_list, ansible,
                external_lb_param, lb_ssh_ip)
        else:
            pass

    def _modify_ansible_user_or_password(self, host_path,
                                         worker_node, ansible):
        try:
            # read hosts.yml
            local_hosts_yaml_path = '/tmp/hosts.yaml'
            self._send_or_receive_file(
                ansible.get('ip_address'), ansible.get('username'),
                ansible.get('password'), host_path,
                local_hosts_yaml_path, 'receive')
            with open(local_hosts_yaml_path, 'r', encoding='utf-8') as f:
                file_content = f.read()
                hosts_content = yaml.safe_load(file_content)
            worker_nodes = hosts_content['all']['children']['kube_node'][
                'hosts']
            LOG.debug("worker_nodes: {}".format(worker_nodes))
            if worker_nodes:
                kube_node_hosts_keys = list(worker_nodes.keys())
                LOG.debug("kube_node_keys: {}".format(kube_node_hosts_keys))
                hosts_key = list(hosts_content['all']['hosts'].keys())
                LOG.debug("hosts_key: {}".format(hosts_key))
                need_modify = False
                for kube_node_hosts in kube_node_hosts_keys:
                    if kube_node_hosts in hosts_key:
                        content = \
                            hosts_content['all']['hosts'][kube_node_hosts]
                        LOG.debug("get node content: {}".format(content))
                        ansible_user = content.get("ansible_user")
                        ansible_password = content.get("ansible_password")
                        if ansible_user != worker_node.get('username'):
                            hosts_content['all']['hosts'][kube_node_hosts][
                                'ansible_user'] = worker_node.get('username')
                            need_modify = True
                        if ansible_password != worker_node.get('password'):
                            hosts_content['all']['hosts'][kube_node_hosts][
                                'ansible_password'] = \
                                worker_node.get('password')
                            need_modify = True
                if need_modify:
                    with open(local_hosts_yaml_path, 'w', encoding='utf-8') \
                            as nf:
                        yaml.safe_dump(hosts_content, nf,
                                       default_flow_style=False)
                    self._send_or_receive_file(
                        ansible.get('ip_address'), ansible.get('username'),
                        ansible.get('password'), host_path,
                        local_hosts_yaml_path, 'send')
                return hosts_content
            os.remove(local_hosts_yaml_path)
        except Exception:
            LOG.error('modify ansible_user or ansible_password has error: {}.'
                      .format(ValueError))
            raise exceptions.MgmtDriverOtherError(
                error_message='modify user or password has error: {}.'.format(
                    Exception))

    def _get_vnfc_resource_id(self, vnfc_resource_list, vnfc_instance_id):
        for vnfc_resource in vnfc_resource_list:
            if vnfc_resource.id == vnfc_instance_id:
                return vnfc_resource
        return None

    def _get_heal_physical_resource_ids(self, vnf_instance,
                                        heal_vnf_request):
        heal_physical_resource_ids = []
        for vnfc_instance_id in heal_vnf_request.vnfc_instance_id:
            instantiated_vnf_info = vnf_instance.instantiated_vnf_info
            vnfc_resource_info = instantiated_vnf_info.vnfc_resource_info
            vnfc_resource = self._get_vnfc_resource_id(
                vnfc_resource_info, vnfc_instance_id)
            if vnfc_resource:
                heal_physical_resource_ids.append(
                    vnfc_resource.compute_resource.resource_id)

        return heal_physical_resource_ids

    def _get_heal_worker_node_info(
            self, vnf_additional_params, worker_node, heatclient,
            nest_stack_id, heal_physical_resource_ids):
        worker_ip_dict_list = []
        if worker_node.get('aspect_id'):
            worker_group_resource_name = worker_node.get('aspect_id')

            worker_group_resource = heatclient.resources.get(
                stack_id=nest_stack_id,
                resource_name=worker_group_resource_name)
            if not worker_group_resource:
                raise exceptions.MgmtDriverOtherError(
                    error_message='The specified resource'
                                  ' {} was not found.'.format(
                                      worker_group_resource_name))
            worker_group_resource_list = heatclient.resources.list(
                stack_id=worker_group_resource.physical_resource_id)
            for worker_resource in worker_group_resource_list:
                lowest_resource_list = heatclient.resources.list(
                    stack_id=worker_resource.physical_resource_id)
                for lowest_resource in lowest_resource_list:
                    if lowest_resource.resource_type == 'OS::Nova::Server' \
                        and lowest_resource.physical_resource_id in \
                            heal_physical_resource_ids:
                        worker_ssh_ip, worker_nic_ip = \
                            self._get_ssh_ip_and_nic_ip(
                                heatclient,
                                worker_resource.physical_resource_id,
                                worker_node)
                        ip_dict = {"nic_ip": worker_nic_ip,
                                   "ssh_ip": worker_ssh_ip}
                        worker_ip_dict_list.append(ip_dict)
        else:
            # in case of SOL001 TOSCA-based VNFD with single worker node
            resource_list = heatclient.resources.list(
                stack_id=nest_stack_id)
            for resource in resource_list:
                if resource.resource_type == 'OS::Nova::Server' \
                    and resource.physical_resource_id in \
                        heal_physical_resource_ids:
                    worker_ssh_ip, worker_nic_ip = \
                        self._get_ssh_ip_and_nic_ip(
                            heatclient, nest_stack_id, worker_node)
                    ip_dict = {"nic_ip": worker_nic_ip,
                               "ssh_ip": worker_ssh_ip}
                    worker_ip_dict_list.append(ip_dict)

        # Get the hostname of the deleting worker nodes
        worker_hostnames = []
        for worker_ip_dict in worker_ip_dict_list:
            # get worker host names
            worker_hostname = \
                'worker' + worker_ip_dict.get("nic_ip").split('.')[-1]
            worker_hostnames.append(worker_hostname)

        return worker_hostnames, worker_ip_dict_list

    @log.log
    def heal_start(self, context, vnf_instance,
                   heal_vnf_request, grant,
                   grant_request, **kwargs):
        vnf_additional_params = \
            vnf_instance.instantiated_vnf_info.additional_params

        # heal of the entire VNF
        if not heal_vnf_request.vnfc_instance_id:
            self.terminate_end(context, vnf_instance, heal_vnf_request,
                               grant, grant_request)
        else:
            # heal specified with VNFC instances
            heal_physical_resource_ids = \
                self._get_heal_physical_resource_ids(
                    vnf_instance, heal_vnf_request)

            worker_node, external_lb_param, ansible, vim_connection_info = \
                self._get_initial_parameters(
                    context, vnf_instance, heal_vnf_request)

            nest_stack_id = vnf_instance.instantiated_vnf_info.instance_id

            # Get the ssh ip of LB
            heatclient = hc.HeatClient(vim_connection_info.access_info)
            ssh_ip, _ = self._get_ssh_ip_and_nic_ip(
                heatclient, nest_stack_id, external_lb_param)

            # Get the worker_hostnames to be healed
            worker_hostnames, _ = self._get_heal_worker_node_info(
                vnf_additional_params, worker_node, heatclient,
                nest_stack_id, heal_physical_resource_ids)

            # remove_worker_node_from_k8s_cluster and update configuration file
            self._remove_node_and_update_config_file(
                worker_hostnames, external_lb_param,
                ssh_ip, ansible, worker_node, 'HEAL')

    @log.log
    def heal_end(self, context, vnf_instance,
                 heal_vnf_request, grant,
                 grant_request, **kwargs):
        vnf_additional_params = \
            vnf_instance.instantiated_vnf_info.additional_params
        # heal of the entire VNF
        if not heal_vnf_request.vnfc_instance_id:
            add_param_list = ['master_node', 'worker_node', 'proxy',
                              'ansible', 'external_lb_param']
            for add_param in add_param_list:
                if heal_vnf_request.additional_params.get(
                        'k8s_cluster_installation_param'):
                    if add_param in heal_vnf_request.additional_params.get(
                            'k8s_cluster_installation_param'):
                        vnf_additional_params.get(
                            'k8s_cluster_installation_param')[add_param] = \
                            heal_vnf_request.additional_params[
                                'k8s_cluster_installation_param'].get(
                                add_param)
            heal_vnf_request.additional_params = vnf_additional_params
            self.instantiate_end(context, vnf_instance, heal_vnf_request,
                                 grant, grant_request)
        else:
            # heal specified with VNFC instances
            heal_physical_resource_ids = \
                self._get_heal_physical_resource_ids(
                    vnf_instance, heal_vnf_request)

            worker_node, external_lb_param, ansible, vim_connection_info = \
                self._get_initial_parameters(
                    context, vnf_instance, heal_vnf_request)

            nest_stack_id = vnf_instance.instantiated_vnf_info.instance_id

            # Get the ssh ip of LB
            heatclient = hc.HeatClient(vim_connection_info.access_info)
            ssh_ip, _ = self._get_ssh_ip_and_nic_ip(
                heatclient, nest_stack_id, external_lb_param)

            # Get the worker_ip_dict_list to be healed
            _, worker_ip_dict_list = self._get_heal_worker_node_info(
                vnf_additional_params, worker_node, heatclient,
                nest_stack_id, heal_physical_resource_ids)

            # Install worker node and update configuration file
            self._install_node_and_update_config_file(
                worker_node, worker_ip_dict_list, ansible,
                external_lb_param, ssh_ip)

    @log.log
    def change_external_connectivity_start(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass

    @log.log
    def change_external_connectivity_end(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass

    @log.log
    def modify_information_start(self, context, vnf_instance,
                                 modify_vnf_request, **kwargs):
        pass

    @log.log
    def modify_information_end(self, context, vnf_instance,
                               modify_vnf_request, **kwargs):
        pass



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_service import loopingcall

from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import http_client


LOG = logging.getLogger(__name__)


CHECK_INTERVAL = 5


class HeatClient(object):

    def __init__(self, vim_info):
        auth = http_client.KeystonePasswordAuthHandle(
            auth_url=vim_info.interfaceInfo['endpoint'],
            username=vim_info.accessInfo['username'],
            password=vim_info.accessInfo['password'],
            project_name=vim_info.accessInfo['project'],
            user_domain_name=vim_info.accessInfo['userDomain'],
            project_domain_name=vim_info.accessInfo['projectDomain']
        )
        self.client = http_client.HttpClient(auth,
                                             service_type='orchestration')

    def create_stack(self, fields, wait=True):
        path = "stacks"
        resp, body = self.client.do_request(path, "POST",
                expected_status=[201], body=fields)

        if wait:
            self.wait_stack_create(fields["stack_name"])

    def update_stack(self, stack_name, fields, wait=True):
        path = "stacks/{}".format(stack_name)
        resp, body = self.client.do_request(path, "PATCH",
                 expected_status=[202], body=fields)

        if wait:
            self.wait_stack_update(stack_name)

    def delete_stack(self, stack_name, wait=True):
        path = "stacks/{}".format(stack_name)
        resp, body = self.client.do_request(path, "DELETE",
                expected_status=[204, 404])

        if wait:
            self.wait_stack_delete(stack_name)

    def get_status(self, stack_name):
        path = "stacks/{}".format(stack_name)
        resp, body = self.client.do_request(path, "GET",
                expected_status=[200, 404])

        if resp.status_code == 404:
            return None, None

        return (body["stack"]["stack_status"],
                body["stack"]["stack_status_reason"])

    def get_resources(self, stack_name):
        # NOTE: Because it is necessary to get nested stack info, it is
        # necessary to specify 'nested_depth=2'.
        path = "stacks/{}/resources?nested_depth=2".format(stack_name)
        resp, body = self.client.do_request(path, "GET",
                expected_status=[200])

        return body['resources']

    def _wait_completion(self, stack_name, operation, complete_status,
            progress_status, failed_status):
        # NOTE: timeout is specified for each stack operation. so it is
        # not forever loop.
        def _check_status():
            status, status_reason = self.get_status(stack_name)
            if status in complete_status:
                LOG.info("%s %s done.", operation, stack_name)
                raise loopingcall.LoopingCallDone()
            elif status in failed_status:
                LOG.error("% %s failed.", operation, stack_name)
                sol_title = "%s failed" % operation
                raise sol_ex.StackOperationFailed(sol_title=sol_title,
                                                  sol_detail=status_reason)
            elif status not in progress_status:
                LOG.error("%s %s failed. status: %s", operation,
                          stack_name, status)
                sol_title = "%s failed" % operation
                raise sol_ex.StackOperationFailed(sol_title=sol_title,
                                                  sol_detail='Unknown error')
            LOG.debug("%s %s %s", operation, stack_name, progress_status)

        timer = loopingcall.FixedIntervalLoopingCall(_check_status)
        timer.start(interval=CHECK_INTERVAL).wait()

    def wait_stack_create(self, stack_name):
        self._wait_completion(stack_name, "Stack create",
            ["CREATE_COMPLETE"], ["CREATE_IN_PROGRESS"], ["CREATE_FAILED"])

    def wait_stack_update(self, stack_name):
        self._wait_completion(stack_name, "Stack update",
            ["UPDATE_COMPLETE"], ["UPDATE_IN_PROGRESS"], ["UPDATE_FAILED"])

    def wait_stack_delete(self, stack_name):
        # NOTE: wait until stack is deleted in the DB since it is necessary
        # for some operations (ex. heal-all).
        # It is expected that it takes short time after "DELETE_COMPLETE".
        # So timeout after "DELETE_COMPLETE" is not specified.
        self._wait_completion(stack_name, "Stack delete",
            [None], ["DELETE_IN_PROGRESS", "DELETE_COMPLETE"],
            ["DELETE_FAILED"])

    def get_stack_resource(self, stack_name):
        path = f"stacks/{stack_name}"
        resp, body = self.client.do_request(path, "GET",
                                            expected_status=[200, 404])
        if resp.status_code == 404:
            raise sol_ex.StackOperationFailed
        return body

    def get_resource_info(self, stack_name, stack_id, resource_name):
        path = f"stacks/{stack_name}/{stack_id}/resources/{resource_name}"
        resp, body = self.client.do_request(path, "GET",
                                            expected_status=[200, 404])
        if resp.status_code == 404:
            return resp, None
        return resp, body['resource']

    def get_parameters(self, stack_name):
        path = "stacks/{}".format(stack_name)
        resp, body = self.client.do_request(path, "GET",
                expected_status=[200])

        return body["stack"]["parameters"]

    def mark_unhealthy(self, stack_id, resource_name):
        path = "stacks/{}/resources/{}".format(stack_id, resource_name)
        fields = {
            "mark_unhealthy": True,
            "resource_status_reason": "marked by tacker"
        }
        resp, body = self.client.do_request(path, "PATCH",
                 expected_status=[200], body=fields)

    def get_template(self, stack_name):
        path = "stacks/{}/template".format(stack_name)
        resp, body = self.client.do_request(path, "GET",
                expected_status=[200])

        return body

    def get_files(self, stack_name):
        path = "stacks/{}/files".format(stack_name)
        resp, body = self.client.do_request(path, "GET",
                expected_status=[200])

        return body


def get_reses_by_types(heat_reses, types):
    return [res for res in heat_reses if res['resource_type'] in types]


def get_server_reses(heat_reses):
    return get_reses_by_types(heat_reses, ['OS::Nova::Server'])


def get_network_reses(heat_reses):
    return get_reses_by_types(heat_reses, ['OS::Neutron::Net'])


def get_storage_reses(heat_reses):
    return get_reses_by_types(heat_reses, ['OS::Cinder::Volume'])


def get_port_reses(heat_reses):
    return get_reses_by_types(heat_reses, ['OS::Neutron::Port'])


def get_stack_name(inst):
    return "vnf-" + inst.id


def get_resource_stack_id(heat_res):
    # return the form "stack_name/stack_id"
    for link in heat_res.get('links', []):
        if link['rel'] == 'stack':
            items = link['href'].split('/')
            return "{}/{}".format(items[-2], items[-1])


def get_parent_resource(heat_res, heat_reses):
    parent = heat_res.get('parent_resource')
    if parent:
        for res in heat_reses:
            if res['resource_name'] == parent:
                return res



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from dateutil import parser
import eventlet
import json
import os
import pickle
import subprocess

from oslo_log import log as logging
from oslo_utils import uuidutils

from tacker.sol_refactored.common import config
from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import vnf_instance_utils as inst_utils
from tacker.sol_refactored.infra_drivers.openstack import heat_utils
from tacker.sol_refactored.infra_drivers.openstack import userdata_default
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects.v2 import fields as v2fields


LOG = logging.getLogger(__name__)

CONF = config.CONF

LINK_PORT_PREFIX = 'req-'
CP_INFO_PREFIX = 'cp-'


# Id of the resources in instantiatedVnfInfo related methods.
# NOTE: instantiatedVnfInfo is re-created in each operation.
# Id of the resources in instantiatedVnfInfo is based on
# heat resource-id so that id is not changed at re-creation.
# Some ids are same as heat resource-id and some ids are
# combination of prefix and other ids.
def _make_link_port_id(link_port_id):
    # prepend 'req-' to distinguish from ports which are
    # created by heat.
    return '{}{}'.format(LINK_PORT_PREFIX, link_port_id)


def _is_link_port(link_port_id):
    return link_port_id.startswith(LINK_PORT_PREFIX)


def _make_cp_info_id(link_port_id):
    return '{}{}'.format(CP_INFO_PREFIX, link_port_id)


def _make_combination_id(a, b):
    return '{}-{}'.format(a, b)


class Openstack(object):

    def __init__(self):
        pass

    def instantiate(self, req, inst, grant_req, grant, vnfd):
        # make HOT
        fields = self._make_hot(req, inst, grant_req, grant, vnfd)
        LOG.debug("stack fields: %s", fields)

        # create or update stack
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        heat_client = heat_utils.HeatClient(vim_info)
        stack_name = heat_utils.get_stack_name(inst)
        status, _ = heat_client.get_status(stack_name)
        if status is None:
            fields['stack_name'] = stack_name
            heat_client.create_stack(fields)
        else:
            heat_client.update_stack(stack_name, fields)

        # get stack resource
        heat_reses = heat_client.get_resources(stack_name)

        # make instantiated_vnf_info
        self._make_instantiated_vnf_info(req, inst, grant_req, grant, vnfd,
            heat_reses)

    def instantiate_rollback(self, req, inst, grant_req, grant, vnfd):
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        heat_client = heat_utils.HeatClient(vim_info)
        stack_name = heat_utils.get_stack_name(inst)
        status, _ = heat_client.get_status(stack_name)
        if status is not None:
            heat_client.delete_stack(stack_name)

    def terminate(self, req, inst, grant_req, grant, vnfd):
        if req.terminationType == 'GRACEFUL':
            timeout = CONF.v2_vnfm.default_graceful_termination_timeout
            if req.obj_attr_is_set('gracefulTerminationTimeout'):
                timeout = req.gracefulTerminationTimeout
            eventlet.sleep(timeout)

        # delete stack
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        heat_client = heat_utils.HeatClient(vim_info)
        stack_name = heat_utils.get_stack_name(inst)
        heat_client.delete_stack(stack_name)

    def _update_nfv_dict(self, heat_client, stack_name, fields):
        parameters = heat_client.get_parameters(stack_name)
        LOG.debug("ORIG parameters: %s", parameters)
        # NOTE: parameters['nfv'] is string
        orig_nfv_dict = json.loads(parameters.get('nfv', '{}'))
        if 'nfv' in fields['parameters']:
            fields['parameters']['nfv'] = inst_utils.json_merge_patch(
                orig_nfv_dict, fields['parameters']['nfv'])
        LOG.debug("NEW parameters: %s", fields['parameters'])
        return fields

    def scale(self, req, inst, grant_req, grant, vnfd):
        # make HOT
        fields = self._make_hot(req, inst, grant_req, grant, vnfd)
        LOG.debug("stack fields: %s", fields)

        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        heat_client = heat_utils.HeatClient(vim_info)

        # mark unhealthy to servers to be removed if scale in
        if req.type == 'SCALE_IN':
            vnfc_res_ids = [res_def.resource.resourceId
                            for res_def in grant_req.removeResources
                            if res_def.type == 'COMPUTE']
            for vnfc in inst.instantiatedVnfInfo.vnfcResourceInfo:
                if vnfc.computeResource.resourceId in vnfc_res_ids:
                    if 'parent_stack_id' not in vnfc.metadata:
                        # It means definition of VDU in the BaseHOT
                        # is inappropriate.
                        raise sol_ex.UnexpectedParentResourceDefinition()
                    heat_client.mark_unhealthy(
                        vnfc.metadata['parent_stack_id'],
                        vnfc.metadata['parent_resource_name'])

        # update stack
        stack_name = heat_utils.get_stack_name(inst)
        fields = self._update_nfv_dict(heat_client, stack_name, fields)
        heat_client.update_stack(stack_name, fields)

        # get stack resource
        heat_reses = heat_client.get_resources(stack_name)

        # make instantiated_vnf_info
        self._make_instantiated_vnf_info(req, inst, grant_req, grant, vnfd,
            heat_reses)

    def scale_rollback(self, req, inst, grant_req, grant, vnfd):
        # NOTE: rollback is supported for scale out only
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        heat_client = heat_utils.HeatClient(vim_info)
        stack_name = heat_utils.get_stack_name(inst)
        heat_reses = heat_client.get_resources(stack_name)

        # mark unhealthy to added servers while scale out
        vnfc_ids = [vnfc.id
                    for vnfc in inst.instantiatedVnfInfo.vnfcResourceInfo]
        for res in heat_utils.get_server_reses(heat_reses):
            if res['physical_resource_id'] not in vnfc_ids:
                metadata = self._make_vnfc_metadata(res, heat_reses)
                if 'parent_stack_id' not in metadata:
                    # It means definition of VDU in the BaseHOT
                    # is inappropriate.
                    raise sol_ex.UnexpectedParentResourceDefinition()
                heat_client.mark_unhealthy(
                    metadata['parent_stack_id'],
                    metadata['parent_resource_name'])

        # update (put back) 'desired_capacity' parameter
        fields = self._update_nfv_dict(heat_client, stack_name,
            userdata_default.DefaultUserData.scale_rollback(
                req, inst, grant_req, grant, vnfd.csar_dir))

        heat_client.update_stack(stack_name, fields)

        # NOTE: instantiatedVnfInfo is not necessary to update since it
        # should be same as before scale API started.

    def change_ext_conn(self, req, inst, grant_req, grant, vnfd):
        # make HOT
        fields = self._make_hot(req, inst, grant_req, grant, vnfd)
        LOG.debug("stack fields: %s", fields)

        # update stack
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        heat_client = heat_utils.HeatClient(vim_info)
        stack_name = heat_utils.get_stack_name(inst)
        fields = self._update_nfv_dict(heat_client, stack_name, fields)
        heat_client.update_stack(stack_name, fields)

        # get stack resource
        heat_reses = heat_client.get_resources(stack_name)

        # make instantiated_vnf_info
        self._make_instantiated_vnf_info(req, inst, grant_req, grant, vnfd,
            heat_reses)

    def change_ext_conn_rollback(self, req, inst, grant_req, grant, vnfd):
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        heat_client = heat_utils.HeatClient(vim_info)
        stack_name = heat_utils.get_stack_name(inst)
        fields = self._update_nfv_dict(heat_client, stack_name,
            userdata_default.DefaultUserData.change_ext_conn_rollback(
                req, inst, grant_req, grant, vnfd.csar_dir))
        heat_client.update_stack(stack_name, fields)

        # NOTE: it is necessary to re-create instantiatedVnfInfo because
        # ports may be changed.
        heat_reses = heat_client.get_resources(stack_name)
        self._make_instantiated_vnf_info(req, inst, grant_req, grant, vnfd,
            heat_reses)

    def heal(self, req, inst, grant_req, grant, vnfd):
        # make HOT
        # NOTE: _make_hot() is called as other operations, but it returns
        # empty 'nfv' dict by default. Therefore _update_nfv_dict() returns
        # current heat parameters as is.
        fields = self._make_hot(req, inst, grant_req, grant, vnfd)
        LOG.debug("stack fields: %s", fields)

        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        heat_client = heat_utils.HeatClient(vim_info)
        stack_name = heat_utils.get_stack_name(inst)
        fields = self._update_nfv_dict(heat_client, stack_name, fields)

        # "re_create" is set to True only when SOL003 heal(without
        # vnfcInstanceId) and "all=True" in additionalParams.
        re_create = False
        if (req.obj_attr_is_set('additionalParams') and
                req.additionalParams.get('all', False) and
                not req.obj_attr_is_set('vnfcInstanceId')):
            re_create = True

        if re_create:
            # NOTE: DefaultUserData::heal() don't care about "template" and
            # "files". Get and set current "template" and "files".
            if "template" not in fields:
                fields["template"] = heat_client.get_template(stack_name)
            if "files" not in fields:
                fields["files"] = heat_client.get_files(stack_name)
            fields["stack_name"] = stack_name

            # stack delete and create
            heat_client.delete_stack(stack_name)
            heat_client.create_stack(fields)
        else:
            # mark unhealthy to target resources.
            # As the target resources has been already selected in
            # constructing grant_req, use it here.
            inst_info = inst.instantiatedVnfInfo

            vnfc_res_ids = [res_def.resource.resourceId
                            for res_def in grant_req.removeResources
                            if res_def.type == 'COMPUTE']
            for vnfc in inst_info.vnfcResourceInfo:
                if vnfc.computeResource.resourceId in vnfc_res_ids:
                    heat_client.mark_unhealthy(
                        vnfc.metadata['stack_id'], vnfc.vduId)

            storage_ids = [res_def.resource.resourceId
                           for res_def in grant_req.removeResources
                           if res_def.type == 'STORAGE']
            if storage_ids:
                for storage_info in inst_info.virtualStorageResourceInfo:
                    if storage_info.storageResource.resourceId in storage_ids:
                        heat_client.mark_unhealthy(
                            storage_info.metadata['stack_id'],
                            storage_info.virtualStorageDescId)

            # update stack
            heat_client.update_stack(stack_name, fields)

        # get stack resource
        heat_reses = heat_client.get_resources(stack_name)

        # make instantiated_vnf_info
        self._make_instantiated_vnf_info(req, inst, grant_req, grant, vnfd,
            heat_reses)

    def _make_hot(self, req, inst, grant_req, grant, vnfd):
        if grant_req.operation == v2fields.LcmOperationType.INSTANTIATE:
            flavour_id = req.flavourId
        else:
            flavour_id = inst.instantiatedVnfInfo.flavourId

        hot_dict = vnfd.get_base_hot(flavour_id)
        if not hot_dict:
            raise sol_ex.BaseHOTNotDefined()

        userdata = None
        userdata_class = None
        if req.obj_attr_is_set('additionalParams'):
            userdata = req.additionalParams.get('lcm-operation-user-data')
            userdata_class = req.additionalParams.get(
                'lcm-operation-user-data-class')

        if userdata is None and userdata_class is None:
            LOG.debug("Processing default userdata %s", grant_req.operation)
            # NOTE: objects used here are dict compat.
            method = getattr(userdata_default.DefaultUserData,
                             grant_req.operation.lower())
            fields = method(req, inst, grant_req, grant, vnfd.csar_dir)
        elif userdata is None or userdata_class is None:
            # Both must be specified.
            raise sol_ex.UserdataMissing()
        else:
            LOG.debug("Processing %s %s %s", userdata, userdata_class,
                grant_req.operation)

            tmp_csar_dir = vnfd.make_tmp_csar_dir()
            script_dict = {
                'request': req.to_dict(),
                'vnf_instance': inst.to_dict(),
                'grant_request': grant_req.to_dict(),
                'grant_response': grant.to_dict(),
                'tmp_csar_dir': tmp_csar_dir
            }
            script_path = os.path.join(
                os.path.dirname(__file__), "userdata_main.py")

            out = subprocess.run(["python3", script_path],
                input=pickle.dumps(script_dict),
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)

            vnfd.remove_tmp_csar_dir(tmp_csar_dir)

            if out.returncode != 0:
                LOG.debug("execute userdata class instantiate failed: %s",
                    out.stderr)
                raise sol_ex.UserdataExecutionFailed(
                    sol_detail=str(out.stderr))

            fields = pickle.loads(out.stdout)

        fields['timeout_mins'] = (
            CONF.v2_vnfm.openstack_vim_stack_create_timeout)

        return fields

    def _get_checked_reses(self, nodes, reses):
        names = list(nodes.keys())

        def _check_res_in_vnfd(res):
            if res['resource_name'] in names:
                return True
            else:
                # should not occur. just check for consistency.
                LOG.debug("%s not in VNFD definition.", res['resource_name'])
                return False

        return {res['physical_resource_id']: res
                for res in reses if _check_res_in_vnfd(res)}

    def _address_range_data_to_info(self, range_data):
        obj = objects.ipOverEthernetAddressInfoV2_IpAddresses_AddressRange()
        obj.minAddress = range_data.minAddress
        obj.maxAddress = range_data.maxAddress
        return obj

    def _proto_data_to_info(self, proto_data):
        # make CpProtocolInfo (5.5.3.9b) from CpProtocolData (4.4.1.10b)
        proto_info = objects.CpProtocolInfoV2(
            layerProtocol=proto_data.layerProtocol
        )
        ip_info = objects.IpOverEthernetAddressInfoV2()

        ip_data = proto_data.ipOverEthernet
        if ip_data.obj_attr_is_set('macAddress'):
            ip_info.macAddress = ip_data.macAddress
        if ip_data.obj_attr_is_set('segmentationId'):
            ip_info.segmentationId = ip_data.segmentationId
        if ip_data.obj_attr_is_set('ipAddresses'):
            addr_infos = []
            for addr_data in ip_data.ipAddresses:
                addr_info = objects.IpOverEthernetAddressInfoV2_IpAddresses(
                    type=addr_data.type)
                if addr_data.obj_attr_is_set('fixedAddresses'):
                    addr_info.addresses = addr_data.fixedAddresses
                if addr_data.obj_attr_is_set('numDynamicAddresses'):
                    addr_info.isDynamic = True
                if addr_data.obj_attr_is_set('addressRange'):
                    addr_info.addressRange = self._address_range_data_to_info(
                        addr_data.addressRange)
                if addr_data.obj_attr_is_set('subnetId'):
                    addr_info.subnetId = addr_data.subnetId
                addr_infos.append(addr_info)
            ip_info.ipAddresses = addr_infos

        proto_info.ipOverEthernet = ip_info

        return proto_info

    def _make_link_ports(self, req_ext_vl, ext_cp_infos):
        link_ports = []
        if not req_ext_vl.obj_attr_is_set('extLinkPorts'):
            return link_ports

        for req_link_port in req_ext_vl.extLinkPorts:
            link_port = objects.ExtLinkPortInfoV2(
                id=_make_link_port_id(req_link_port.id),
                resourceHandle=req_link_port.resourceHandle,
            )
            ext_cp_info = objects.VnfExtCpInfoV2(
                id=_make_cp_info_id(link_port.id),
                extLinkPortId=link_port.id
                # associatedVnfcCpId may set later
            )
            link_port.cpInstanceId = ext_cp_info.id

            for ext_cp in req_ext_vl.extCps:
                found = False
                for key, cp_conf in ext_cp.cpConfig.items():
                    if (cp_conf.obj_attr_is_set('linkPortId') and
                            cp_conf.linkPortId == req_link_port.id):
                        ext_cp_info.cpdId = ext_cp.cpdId
                        ext_cp_info.cpConfigId = key
                        # NOTE: cpProtocolInfo can't be filled
                        found = True
                        break
                if found:
                    break

            link_ports.append(link_port)
            ext_cp_infos.append(ext_cp_info)

        return link_ports

    def _make_ext_vl_from_req(self, req_ext_vl, ext_cp_infos):
        ext_vl = objects.ExtVirtualLinkInfoV2(
            id=req_ext_vl.id,
            resourceHandle=objects.ResourceHandle(
                resourceId=req_ext_vl.resourceId
            ),
            currentVnfExtCpData=req_ext_vl.extCps
        )
        if req_ext_vl.obj_attr_is_set('vimConnectionId'):
            ext_vl.resourceHandle.vimConnectionId = (
                req_ext_vl.vimConnectionId)
        if req_ext_vl.obj_attr_is_set('resourceProviderId'):
            ext_vl.resourceHandle.resourceProviderId = (
                req_ext_vl.resourceProviderId)

        link_ports = self._make_link_ports(req_ext_vl, ext_cp_infos)
        if link_ports:
            ext_vl.extLinkPorts = link_ports

        ext_vl.extLinkPorts = link_ports

        return ext_vl

    def _make_ext_vl_info_from_req(self, req, grant, ext_cp_infos):
        # make extVirtualLinkInfo
        req_ext_vls = []
        if grant.obj_attr_is_set('extVirtualLinks'):
            req_ext_vls = grant.extVirtualLinks
        elif req.obj_attr_is_set('extVirtualLinks'):
            req_ext_vls = req.extVirtualLinks

        return [self._make_ext_vl_from_req(req_ext_vl, ext_cp_infos)
                for req_ext_vl in req_ext_vls]

    def _find_ext_cp_info(self, link_port, ext_cp_infos):
        for ext_cp in ext_cp_infos:
            if ext_cp.id == link_port.cpInstanceId:
                return ext_cp
        # never reach here

    def _make_ext_vl_info_from_inst(self, old_inst_vnf_info, ext_cp_infos):
        # make extVirtualLinkInfo from old inst.extVirtualLinkInfo
        ext_vls = []
        old_cp_infos = []

        if old_inst_vnf_info.obj_attr_is_set('extVirtualLinkInfo'):
            ext_vls = old_inst_vnf_info.extVirtualLinkInfo
        if old_inst_vnf_info.obj_attr_is_set('extCpInfo'):
            old_cp_infos = old_inst_vnf_info.extCpInfo

        for ext_vl in ext_vls:
            if not ext_vl.obj_attr_is_set('extLinkPorts'):
                continue
            new_link_ports = []
            for link_port in ext_vl.extLinkPorts:
                if _is_link_port(link_port.id):
                    new_link_ports.append(link_port)
                    ext_cp_infos.append(self._find_ext_cp_info(link_port,
                                                               old_cp_infos))
            ext_vl.extLinkPorts = new_link_ports

        return ext_vls

    def _make_ext_vl_info_from_req_and_inst(self, req, grant, old_inst_info,
            ext_cp_infos):
        req_ext_vls = []
        if grant.obj_attr_is_set('extVirtualLinks'):
            req_ext_vls = grant.extVirtualLinks
        elif req.obj_attr_is_set('extVirtualLinks'):
            req_ext_vls = req.extVirtualLinks

        req_ext_vl_ids = {ext_vl.id for ext_vl in req_ext_vls}
        inst_ext_vl_ids = set()
        if old_inst_info.obj_attr_is_set('extVirtualLinkInfo'):
            inst_ext_vl_ids = {ext_vl.id
                               for ext_vl in old_inst_info.extVirtualLinkInfo}

        added_ext_vl_ids = req_ext_vl_ids - inst_ext_vl_ids
        req_all_cp_names = {ext_cp.cpdId
                            for req_ext_vl in req_ext_vls
                            for ext_cp in req_ext_vl.extCps}

        ext_vls = [self._make_ext_vl_from_req(req_ext_vl, ext_cp_infos)
                   for req_ext_vl in req_ext_vls
                   # added ext_vls
                   if req_ext_vl.id in added_ext_vl_ids]

        old_ext_vls = []
        old_cp_infos = []
        if old_inst_info.obj_attr_is_set('extVirtualLinkInfo'):
            old_ext_vls = old_inst_info.extVirtualLinkInfo
        if old_inst_info.obj_attr_is_set('extCpInfo'):
            old_cp_infos = old_inst_info.extCpInfo

        for ext_vl in old_ext_vls:
            old_ext_cp_data = ext_vl.currentVnfExtCpData
            old_link_ports = (ext_vl.extLinkPorts
                if ext_vl.obj_attr_is_set('extLinkPorts') else [])
            new_ext_cp_data = []
            new_link_ports = []

            for ext_cp_data in old_ext_cp_data:
                if ext_cp_data.cpdId not in req_all_cp_names:
                    new_ext_cp_data.append(ext_cp_data)

            for link_port in old_link_ports:
                ext_cp = self._find_ext_cp_info(link_port, old_cp_infos)
                if (ext_cp.cpdId not in req_all_cp_names and
                        _is_link_port(link_port.id)):
                    new_link_ports.append(link_port)
                    ext_cp_infos.append(ext_cp)

            def _find_req_ext_vl(ext_vl):
                for req_ext_vl in req_ext_vls:
                    if req_ext_vl.id == ext_vl.id:
                        return req_ext_vl

            req_ext_vl = _find_req_ext_vl(ext_vl)
            if req_ext_vl is not None:
                new_ext_cp_data += req_ext_vl.extCps
                new_link_ports += self._make_link_ports(req_ext_vl,
                                                        ext_cp_infos)

            if new_ext_cp_data:
                # if it is empty, it means all cps of this ext_vl are replaced
                # by another ext_vl.
                ext_vl.currentVnfExtCpData = new_ext_cp_data
                ext_vl.extLinkPorts = new_link_ports
                ext_vls.append(ext_vl)

        return ext_vls

    def _make_ext_mgd_vl_info_from_req(self, vnfd, flavour_id, req, grant):
        # make extManagedVirtualLinkInfo
        ext_mgd_vls = []
        req_mgd_vls = []
        if grant.obj_attr_is_set('extManagedVirtualLinks'):
            req_mgd_vls = grant.extManagedVirtualLinks
        elif req.obj_attr_is_set('extManagedVirtualLinks'):
            req_mgd_vls = req.extManagedVirtualLinks

        vls = vnfd.get_virtual_link_nodes(flavour_id)
        for req_mgd_vl in req_mgd_vls:
            vl_name = req_mgd_vl.vnfVirtualLinkDescId
            if vl_name not in list(vls.keys()):
                # should not occur. just check for consistency.
                LOG.debug("%s not in VNFD VL definition.", vl_name)
                continue
            ext_mgd_vl = objects.ExtManagedVirtualLinkInfoV2(
                id=req_mgd_vl.id,
                vnfVirtualLinkDescId=vl_name,
                networkResource=objects.ResourceHandle(
                    id=uuidutils.generate_uuid(),
                    resourceId=req_mgd_vl.resourceId
                )
            )
            if req_mgd_vl.obj_attr_is_set('vimConnectionId'):
                ext_mgd_vl.networkResource.vimConnectionId = (
                    req_mgd_vl.vimConnectionId)
            if req_mgd_vl.obj_attr_is_set('resourceProviderId'):
                ext_mgd_vl.networkResource.resourceProviderId = (
                    req_mgd_vl.resourceProviderId)

            ext_mgd_vls.append(ext_mgd_vl)

            if req_mgd_vl.obj_attr_is_set('vnfLinkPort'):
                ext_mgd_vl.vnfLinkPort = [
                    objects.VnfLinkPortInfoV2(
                        id=_make_link_port_id(req_link_port.vnfLinkPortId),
                        resourceHandle=req_link_port.resourceHandle,
                        cpInstanceType='EXT_CP',  # may be changed later
                        # cpInstanceId may set later
                    )
                    for req_link_port in req_mgd_vl.vnfLinkPort
                ]

        return ext_mgd_vls

    def _make_ext_mgd_vl_info_from_inst(self, old_inst_vnf_info):
        # make extManagedVirtualLinkInfo
        ext_mgd_vls = []

        if old_inst_vnf_info.obj_attr_is_set('extManagedVirtualLinkInfo'):
            ext_mgd_vls = old_inst_vnf_info.extManagedVirtualLinkInfo

        for ext_mgd_vl in ext_mgd_vls:
            if ext_mgd_vl.obj_attr_is_set('vnfLinkPorts'):
                ext_mgd_vl.vnfLinkPorts = [link_port
                    for link_port in ext_mgd_vl.vnfLinkPorts
                    if _is_link_port(link_port.id)]

        return ext_mgd_vls

    def _find_ext_vl_by_cp_name(self, cp_name, ext_vl_infos):
        for ext_vl_info in ext_vl_infos:
            for ext_cp_data in ext_vl_info.currentVnfExtCpData:
                if ext_cp_data.cpdId == cp_name:
                    return ext_vl_info, ext_cp_data

        return None, None

    def _link_ext_port_info(self, ext_port_infos, ext_vl_infos, ext_cp_infos,
            port_reses):
        for ext_port_info in ext_port_infos:
            res = port_reses[ext_port_info.id]
            cp_name = res['resource_name']
            ext_cp_info = objects.VnfExtCpInfoV2(
                id=_make_cp_info_id(ext_port_info.id),
                extLinkPortId=ext_port_info.id,
                cpdId=cp_name
                # associatedVnfcCpId may set later
            )
            ext_port_info.cpInstanceId = ext_cp_info.id

            ext_vl_info, ext_cp_data = self._find_ext_vl_by_cp_name(
                cp_name, ext_vl_infos)

            if ext_vl_info:
                if ext_vl_info.obj_attr_is_set('extLinkPorts'):
                    ext_vl_info.extLinkPorts.append(ext_port_info)
                else:
                    ext_vl_info.extLinkPorts = [ext_port_info]

                for key, cp_conf in ext_cp_data.cpConfig.items():
                    # NOTE: it is assumed that there is one item
                    # (with cpProtocolData) of cpConfig at the moment.
                    if cp_conf.obj_attr_is_set('cpProtocolData'):
                        proto_infos = []
                        for proto_data in cp_conf.cpProtocolData:
                            proto_info = self._proto_data_to_info(
                                proto_data)
                            proto_infos.append(proto_info)
                        ext_cp_info.cpProtocolInfo = proto_infos
                        ext_cp_info.cpConfigId = key
                        break

            ext_cp_infos.append(ext_cp_info)

    def _find_vnfc_cp_info(self, port_res, vnfc_res_infos, server_reses):
        for vnfc_res_info in vnfc_res_infos:
            if not vnfc_res_info.obj_attr_is_set('vnfcCpInfo'):
                continue
            vnfc_res = server_reses[vnfc_res_info.id]
            vdu_name = vnfc_res_info.vduId
            cp_name = port_res['resource_name']
            if (vdu_name in port_res.get('required_by', []) and
                    port_res.get('parent_resource') ==
                    vnfc_res.get('parent_resource')):
                for vnfc_cp in vnfc_res_info.vnfcCpInfo:
                    if vnfc_cp.cpdId == cp_name:
                        return vnfc_cp

    def _link_vnfc_cp_info(self, vnfc_res_infos, ext_port_infos,
            vnf_port_infos, ext_cp_infos, server_reses, port_reses):

        for ext_port_info in ext_port_infos:
            port_res = port_reses[ext_port_info.id]
            vnfc_cp = self._find_vnfc_cp_info(port_res, vnfc_res_infos,
                server_reses)
            if vnfc_cp:
                # should be found
                vnfc_cp.vnfExtCpId = ext_port_info.cpInstanceId
                for ext_cp_info in ext_cp_infos:
                    if ext_cp_info.extLinkPortId == ext_port_info.id:
                        ext_cp_info.associatedVnfcCpId = vnfc_cp.id
                        break

        for vnf_port_info in vnf_port_infos:
            port_res = port_reses[vnf_port_info.id]
            vnfc_cp = self._find_vnfc_cp_info(port_res, vnfc_res_infos,
                server_reses)
            if vnfc_cp:
                # should be found
                vnf_port_info.cpInstanceType = 'VNFC_CP'
                vnf_port_info.cpInstanceId = vnfc_cp.id
                vnfc_cp.vnfLinkPortId = vnf_port_info.id

    def _make_vnfc_metadata(self, server_res, heat_reses):
        metadata = {
            'creation_time': server_res['creation_time'],
            'stack_id': heat_utils.get_resource_stack_id(server_res)
        }
        parent_res = heat_utils.get_parent_resource(server_res, heat_reses)
        if parent_res:
            metadata['parent_stack_id'] = (
                heat_utils.get_resource_stack_id(parent_res))
            metadata['parent_resource_name'] = parent_res['resource_name']

        return metadata

    def _make_instantiated_vnf_info(self, req, inst, grant_req, grant, vnfd,
            heat_reses):
        op = grant_req.operation
        if op == v2fields.LcmOperationType.INSTANTIATE:
            flavour_id = req.flavourId
        else:
            flavour_id = inst.instantiatedVnfInfo.flavourId
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        vducp_nodes = vnfd.get_vducp_nodes(flavour_id)

        storage_reses = self._get_checked_reses(
            vnfd.get_storage_nodes(flavour_id),
            heat_utils.get_storage_reses(heat_reses))
        server_reses = self._get_checked_reses(vnfd.get_vdu_nodes(flavour_id),
            heat_utils.get_server_reses(heat_reses))
        network_reses = self._get_checked_reses(
            vnfd.get_virtual_link_nodes(flavour_id),
            heat_utils.get_network_reses(heat_reses))
        port_reses = self._get_checked_reses(vducp_nodes,
            heat_utils.get_port_reses(heat_reses))

        def _res_to_handle(res):
            return objects.ResourceHandle(
                resourceId=res['physical_resource_id'],
                vimLevelResourceType=res['resource_type'],
                vimConnectionId=vim_info.vimId)

        storage_infos = [
            objects.VirtualStorageResourceInfoV2(
                id=res_id,
                virtualStorageDescId=res['resource_name'],
                storageResource=_res_to_handle(res),
                metadata={
                    'stack_id': heat_utils.get_resource_stack_id(res)
                }
            )
            for res_id, res in storage_reses.items()
        ]

        vnfc_res_infos = [
            objects.VnfcResourceInfoV2(
                id=res_id,
                vduId=res['resource_name'],
                computeResource=_res_to_handle(res),
                metadata=self._make_vnfc_metadata(res, heat_reses)
            )
            for res_id, res in server_reses.items()
        ]

        for vnfc_res_info in vnfc_res_infos:
            vdu_name = vnfc_res_info.vduId
            server_res = server_reses[vnfc_res_info.id]
            storage_ids = [storage_id
                for storage_id, storage_res in storage_reses.items()
                if (vdu_name in storage_res.get('required_by', []) and
                    server_res.get('parent_resource') ==
                    storage_res.get('parent_resource'))
            ]
            if storage_ids:
                vnfc_res_info.storageResourceIds = storage_ids

            vdu_cps = vnfd.get_vdu_cps(flavour_id, vdu_name)
            cp_infos = [
                objects.VnfcResourceInfoV2_VnfcCpInfo(
                    id=_make_combination_id(cp, vnfc_res_info.id),
                    cpdId=cp,
                    # vnfExtCpId or vnfLinkPortId may set later
                )
                for cp in vdu_cps
            ]
            if cp_infos:
                vnfc_res_info.vnfcCpInfo = cp_infos

        vnf_vl_res_infos = [
            objects.VnfVirtualLinkResourceInfoV2(
                id=res_id,
                vnfVirtualLinkDescId=res['resource_name'],
                networkResource=_res_to_handle(res)
            )
            for res_id, res in network_reses.items()
        ]

        ext_cp_infos = []
        if op == v2fields.LcmOperationType.INSTANTIATE:
            ext_vl_infos = self._make_ext_vl_info_from_req(
                req, grant, ext_cp_infos)
            ext_mgd_vl_infos = self._make_ext_mgd_vl_info_from_req(vnfd,
                flavour_id, req, grant)
        else:
            old_inst_vnf_info = inst.instantiatedVnfInfo
            if op == v2fields.LcmOperationType.CHANGE_EXT_CONN:
                ext_vl_infos = self._make_ext_vl_info_from_req_and_inst(
                    req, grant, old_inst_vnf_info, ext_cp_infos)
            else:
                ext_vl_infos = self._make_ext_vl_info_from_inst(
                    old_inst_vnf_info, ext_cp_infos)
            ext_mgd_vl_infos = self._make_ext_mgd_vl_info_from_inst(
                old_inst_vnf_info)

        def _find_vl_name(port_res):
            cp_name = port_res['resource_name']
            return vnfd.get_vl_name_from_cp(flavour_id, vducp_nodes[cp_name])

        ext_port_infos = [
            objects.ExtLinkPortInfoV2(
                id=res_id,
                resourceHandle=_res_to_handle(res)
            )
            for res_id, res in port_reses.items()
            if _find_vl_name(res) is None
        ]

        self._link_ext_port_info(ext_port_infos, ext_vl_infos, ext_cp_infos,
            port_reses)

        vnf_port_infos = [
            objects.VnfLinkPortInfoV2(
                id=res_id,
                resourceHandle=_res_to_handle(res),
                cpInstanceType='EXT_CP'  # may be changed later
            )
            for res_id, res in port_reses.items()
            if _find_vl_name(res) is not None
        ]

        vl_name_to_info = {info.vnfVirtualLinkDescId: info
            for info in vnf_vl_res_infos + ext_mgd_vl_infos}

        for vnf_port_info in vnf_port_infos:
            port_res = port_reses[vnf_port_info.id]
            vl_info = vl_name_to_info.get(_find_vl_name(port_res))
            if vl_info is None:
                # should not occur. just check for consistency.
                continue

            if vl_info.obj_attr_is_set('vnfLinkPorts'):
                vl_info.vnfLinkPorts.append(vnf_port_info)
            else:
                vl_info.vnfLinkPorts = [vnf_port_info]

        self._link_vnfc_cp_info(vnfc_res_infos, ext_port_infos,
            vnf_port_infos, ext_cp_infos, server_reses, port_reses)

        # NOTE: The followings are not handled at the moment.
        # - handle tosca.nodes.nfv.VnfExtCp type
        #   Note that there is no example in current tacker examples which use
        #   tosca.nodes.nfv.VnfExtCp type and related BaseHOT definitions.
        # - in the case of specifying linkPortId of extVirtualLinks or
        #   extManagedVirtualLinks, the link of vnfcCpInfo is not handled
        #   because the association of compute resource and port resource
        #   is not identified.

        # make new instantiatedVnfInfo and replace
        inst_vnf_info = objects.VnfInstanceV2_InstantiatedVnfInfo(
            flavourId=flavour_id,
            vnfState='STARTED',
        )
        if storage_infos:
            inst_vnf_info.virtualStorageResourceInfo = storage_infos

        if vnfc_res_infos:
            # NOTE: scale-in specification of tacker SOL003 v2 API is that
            # newer VDU is selected for reduction. It is necessary to sort
            # vnfc_res_infos at this point so that the conductor should
            # choose VDUs from a head sequentially when making scale-in
            # grant request.

            def _get_key(vnfc):
                return parser.isoparse(vnfc.metadata['creation_time'])

            sorted_vnfc_res_infos = sorted(vnfc_res_infos, key=_get_key,
                                           reverse=True)
            inst_vnf_info.vnfcResourceInfo = sorted_vnfc_res_infos

        if vnf_vl_res_infos:
            inst_vnf_info.vnfVirtualLinkResourceInfo = vnf_vl_res_infos

        if ext_vl_infos:
            inst_vnf_info.extVirtualLinkInfo = ext_vl_infos

        if ext_mgd_vl_infos:
            inst_vnf_info.extManagedVirtualLinkInfo = ext_mgd_vl_infos

        if ext_cp_infos:
            inst_vnf_info.extCpInfo = ext_cp_infos

        # make vnfcInfo
        # NOTE: vnfcInfo only exists in SOL002
        if vnfc_res_infos:
            vnfc_infos = []
            old_vnfc_infos = []
            if (inst.obj_attr_is_set('instantiatedVnfInfo') and
                    inst.instantiatedVnfInfo.obj_attr_is_set('vnfcInfo')):
                old_vnfc_infos = inst.instantiatedVnfInfo.vnfcInfo

            for vnfc_res_info in sorted_vnfc_res_infos:
                vnfc_info = None
                vnfc_id = _make_combination_id(vnfc_res_info.vduId,
                                               vnfc_res_info.id)
                for old_vnfc_info in old_vnfc_infos:
                    if old_vnfc_info.id == vnfc_id:
                        # re-use current object since
                        # vnfcConfigurableProperties may be set.
                        vnfc_info = old_vnfc_info
                        break
                if vnfc_info is None:
                    vnfc_info = objects.VnfcInfoV2(
                        id=vnfc_id,
                        vduId=vnfc_res_info.vduId,
                        vnfcResourceInfoId=vnfc_res_info.id,
                        vnfcState='STARTED'
                    )
                vnfc_infos.append(vnfc_info)

            inst_vnf_info.vnfcInfo = vnfc_infos

        inst.instantiatedVnfInfo = inst_vnf_info



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from tacker import context
from tacker import objects
from tacker.tests.unit import base
from tacker.tests import uuidsentinel


class GrantTestCase(base.TestCase):

    def setUp(self):
        super(GrantTestCase, self).setUp()
        self.context = context.get_admin_context()

    def _get_grant_data(self):
        vim_connection_info = [{
            "id": "6b0ff598-60d6-49b4-a907-a1111de52d92",
            "vim_id": uuidsentinel.vim_id,
            "vim_type": "ETSINFV.OPENSTACK_KEYSTONE.v_2",
            "interface_info": {
                "endpoint": "endpoint_value"},
            "access_info": {
                "username": "username_value",
                "password": "password_value",
                "region": "region_value",
                "tenant": "tenant_value"}}]

        grant_info = [{
            "resource_definition_id":
                "ff150558-6f45-11eb-9439-0242ac130002"}, {
            "resource_definition_id":
                "030df048-6f46-11eb-9439-0242ac130002"}]

        vim_compute_resource_flavour = [{
            "vim_connection_id": "6b0ff598-60d6-49b4-a907-a1111de52d92",
            "vnfd_virtual_compute_desc_id":
                "a813a1ea-6f47-11eb-9439-0242ac130002",
            "vim_flavour_id": "c924aa12-b69d-41ee-91dd-7a26d92e6147"}]

        vim_software_image = [{
            "vim_connection_id": "6b0ff598-60d6-49b4-a907-a1111de52d92",
            "vnfd_software_image_id": "49597d31-0aad-4c46-865d-ed7e60b3edc7",
            "vim_software_image_id": "93738ce0-3ddb-4c13-9900-62b757561edd"}]

        ext_vl_info = [{
            "id": "external_network",
            "vim_connection_id": "6b0ff598-60d6-49b4-a907-a1111de52d92",
            "resource_id": "dc67ee99-e963-44e2-a152-f0fb492eae76",
            "ext_cps": [{
                "cpd_id": "CP1",
                "cp_config": [{
                    "cp_protocol_data": [{
                        "ip_over_ethernet": {
                            "mac_address": "fa:16:3e:0d:6f:71"},
                        "layer_protocol": "IP_OVER_ETHERNET"}]}]}, {
                "cpd_id": "CP2",
                "cp_config": [{
                    "cp_protocol_data": [{
                        "ip_over_ethernet": {
                            "ipaddresses": [{
                                "type": "IPV4",
                                "fixed_addresses": [
                                    "10.0.0.1"],
                                "subnet_id":
                                    "55f0fb3c-6a70-11eb-9439-0242ac130002"}]},
                        "layer_protocol": "IP_OVER_ETHERNET"}]}]}, {
                "cpd_id": "CP3",
                "cp_config": [{
                    "cp_protocol_data": [{
                        "ip_over_ethernet": {
                            "ipaddresses": [{
                                "type": "IPV4",
                                "num_dynamic_addresses": 1,
                                "subnet_id":
                                    "3a7a37fc-6a92-11eb-9439-0242ac130002"}]},
                        "layer_protocol": "IP_OVER_ETHERNET"}]}]}, {
                "cpd_id": "CP4",
                "cp_config": [{
                    "cp_protocol_data": [{
                        "layer_protocol": "IP_OVER_ETHERNET"}],
                    "link_port_id":
                        "413f4e46-21cf-41b1-be0f-de8d23f76cfe"}]}],
            "ext_link_ports": [{
                "id": "413f4e46-21cf-41b1-be0f-de8d23f76cfe",
                "resource_handle": {
                    "resource_id": "67f7e772-0d31-4087-bf4c-2576fadcbdb7",
                    "vim_connection_id":
                        "6b0ff598-60d6-49b4-a907-a1111de52d92",
                    "vim_level_resource_type": "LINKPORT"}}], }]

        grant_data = {
            "id": "3affab40-6f28-11eb-9439-0242ac130002",
            "vnf_instance_id": "4d62c7c2-6f28-11eb-9439-0242ac130002",
            "vnf_lcm_op_occ_id": "6e426236-6f28-11eb-9439-0242ac130002",
            "vim_connections": vim_connection_info,
            "update_resources": grant_info,
            "vim_assets": {
                "compute_resource_flavours": vim_compute_resource_flavour,
                "software_images": vim_software_image,
            },
            "ext_virtual_links": ext_vl_info,
            "additional_params": {"key1": "value1"},
            "_links": {
                "self": {"href": ""},
                "vnf_lcm_op_occ": {"href": ""},
                "vnf_instance": {"href": ""},
            }
        }

        return grant_data

    def test_obj_from_primitive(self):
        grant_data = self._get_grant_data()
        grant = objects.Grant.obj_from_primitive(
            copy.deepcopy(grant_data), self.context)
        self._check_grant(grant, grant_data)

    def _check_grant(self, obj, data):

        def _check_vim_connection_info(_obj, _data):
            self.assertEqual(len(_obj), len(_data))
            for obj, data in zip(_obj, _data):
                self.assertIsInstance(obj, objects.VimConnectionInfo)

        def _check_update_resources(_obj, _data):
            self.assertEqual(len(_obj), len(_data))
            for obj, data in zip(_obj, _data):
                self.assertIsInstance(obj, objects.GrantInfo)

        def _check_vim_assets(obj, data):
            self.assertIsInstance(obj, objects.VimAssets)
            _check_compute_resource_flavours(
                obj.compute_resource_flavours,
                data.get('compute_resource_flavours', []))
            _check_software_images(obj.software_images,
                data.get('software_images', []))

        def _check_compute_resource_flavours(_obj, _data):
            self.assertEqual(len(_obj), len(_data))
            for obj, data in zip(_obj, _data):
                self.assertIsInstance(obj,
                    objects.VimComputeResourceFlavour)
                self.assertEqual(obj.vim_connection_id,
                    data.get('vim_connection_id'))
                self.assertEqual(obj.vnfd_virtual_compute_desc_id,
                    data.get('vnfd_virtual_compute_desc_id'))
                self.assertEqual(obj.vim_flavour_id,
                    data.get('vim_flavour_id'))

        def _check_software_images(_obj, _data):
            self.assertEqual(len(_obj), len(_data))
            for obj, data in zip(_obj, _data):
                self.assertIsInstance(obj, objects.VimSoftwareImage)
                self.assertEqual(obj.vim_connection_id,
                    data.get('vim_connection_id'))
                self.assertEqual(obj.vnfd_software_image_id,
                    data.get('vnfd_software_image_id'))
                self.assertEqual(obj.vim_software_image_id,
                    data.get('vim_software_image_id'))

        def _check_external_virtual_links(_obj, _data):
            self.assertEqual(len(_obj), len(_data))
            for obj, data in zip(_obj, _data):
                self.assertIsInstance(obj, objects.ExtVirtualLinkData)
                self.assertEqual(obj.id, data.get('id'))
                self.assertEqual(obj.vim_connection_id,
                    data.get('vim_connection_id'))
                self.assertEqual(obj.resource_id,
                    data.get('resource_id'))
                _check_ext_cps(obj.ext_cps,
                    data.get('ext_cps', []))
                _check_ext_link_ports(obj.ext_link_ports,
                    data.get('ext_link_ports', []))

        def _check_ext_cps(_obj, _data):
            self.assertEqual(len(_obj), len(_data))
            for obj, data in zip(_obj, _data):
                self.assertIsInstance(obj, objects.VnfExtCpData)
                self.assertEqual(obj.cpd_id, data.get('cpd_id'))
                _check_ext_cp_config(obj.cp_config,
                    data.get('cp_config', []))

        def _check_ext_cp_config(_obj, _data):
            self.assertEqual(len(_obj), len(_data))
            for obj, data in zip(_obj, _data):
                self.assertIsInstance(obj, objects.VnfExtCpConfig)
                self.assertEqual(obj.cp_instance_id,
                    data.get('cp_instance_id'))
                self.assertEqual(obj.link_port_id,
                    data.get('link_port_id'))
                _check_cp_protocol_data(obj.cp_protocol_data,
                    data.get('cp_protocol_data', []))

        def _check_cp_protocol_data(_obj, _data):
            self.assertEqual(len(_obj), len(_data))
            for obj, data in zip(_obj, _data):
                self.assertIsInstance(obj, objects.CpProtocolData)
                self.assertEqual(obj.layer_protocol,
                    data.get('layer_protocol'))
                if obj.ip_over_ethernet or data.get('ip_over_ethernet', None):
                    _check_ip_over_ethernet(obj.ip_over_ethernet,
                        data.get('ip_over_ethernet', None))

        def _check_ip_over_ethernet(obj, data):
            self.assertIsInstance(obj, objects.IpOverEthernetAddressData)
            self.assertEqual(obj.mac_address, data.get('mac_address'))
            _check_ip_addresses(obj.ip_addresses,
                data.get('ip_addresses', []))

        def _check_ip_addresses(_obj, _data):
            self.assertEqual(len(_obj), len(_data))
            for obj, data in zip(_obj, _data):
                self.assertIsInstance(obj, objects.IpAddressReq)
                self.assertEqual(obj.type, data.get('type'))
                self.assertEqual(obj.subnet_id, data.get('subnet_id'))
                self.assertEqual(obj.num_dynamic_addresses,
                    data.get('num_dynamic_addresses'))

        def _check_ext_link_ports(_obj, _data):
            self.assertEqual(len(_obj), len(_data))
            for obj, data in zip(_obj, _data):
                self.assertIsInstance(obj, objects.ExtLinkPortData)
                self.assertEqual(obj.id, data.get('id'))
                self.assertIsInstance(obj.resource_handle,
                    objects.ResourceHandle)

        self.assertIsInstance(obj, objects.Grant)
        self.assertEqual(obj.id, data.get('id'))
        self.assertEqual(obj.vnf_instance_id, data.get('vnf_instance_id'))
        self.assertEqual(obj.vnf_lcm_op_occ_id, data.get('vnf_lcm_op_occ_id'))
        _check_vim_connection_info(obj.vim_connections,
            data.get('vim_connections'))
        _check_update_resources(obj.update_resources,
            data.get('update_resources'))
        _check_vim_assets(obj.vim_assets,
            data.get('vim_assets', None))
        _check_external_virtual_links(obj.ext_virtual_links,
            data.get('ext_virtual_links'))
